"""
A basic class to serve as the document controller of a typical one window application.
"""

# standard libraries
import asyncio
import gettext
import logging
import typing

# local libraries
from nion.utils import Geometry
from nion.utils import Process
from nion.ui import UserInterface


_ = gettext.gettext


class Window:

    def __init__(self, ui: UserInterface.UserInterface, app=None, parent_window=None, window_style=None, persistent_id=None):
        self.ui = ui
        self.app = app
        self.on_close = None
        parent_window = parent_window._document_window if parent_window else None
        self.__document_window = self.ui.create_document_window(parent_window=parent_window)
        if window_style:
            self.__document_window.window_style = window_style
        self.__persistent_id = persistent_id
        self.__shown = False
        self.__document_window.on_periodic = self.periodic
        self.__document_window.on_queue_task = self.queue_task
        self.__document_window.on_clear_queued_tasks = self.clear_queued_tasks
        self.__document_window.on_add_task = self.add_task
        self.__document_window.on_clear_task = self.clear_task
        self.__document_window.on_about_to_show = self.about_to_show
        self.__document_window.on_about_to_close = self.about_to_close
        self.__document_window.on_activation_changed = self.activation_changed
        self.__document_window.on_key_pressed = self.key_pressed
        self.__document_window.on_key_released = self.key_released
        self.__document_window.on_size_changed = self.size_changed
        self.__document_window.on_position_changed = self.position_changed
        self.__document_window.on_refocus_widget = self.refocus_widget
        self.__document_window.on_ui_activity = self._register_ui_activity
        self.__periodic_queue = Process.TaskQueue()
        self.__periodic_set = Process.TaskSet()
        # configure the event loop object
        logger = logging.getLogger()
        old_level = logger.level
        logger.setLevel(logging.INFO)
        self.__event_loop = asyncio.new_event_loop()  # outputs a debugger message!
        logger.setLevel(old_level)

    def close(self):
        self.on_close = None
        Process.close_event_loop(self.__event_loop)
        self.__event_loop = None
        self.ui.destroy_document_window(self.__document_window)  # close the ui window
        self.__document_window = None
        self.__periodic_queue = None
        self.__periodic_set = None

    @property
    def _document_window(self):
        # for testing only
        return self.__document_window

    def _create_menus(self):
        self._file_menu = self.add_menu(_("File"))
        self._edit_menu = self.add_menu(_("Edit"))
        self._window_menu = self.add_menu(_("Window"))
        self._help_menu = self.add_menu(_("Help"))

        self._close_action = self._file_menu.add_menu_item(_("Close Window"), self.request_close, key_sequence="close")
        self._file_menu.add_separator()
        self._page_setup_action = self._file_menu.add_menu_item(_("Page Setup"), self._page_setup)
        self._print_action = self._file_menu.add_menu_item(_("Print"), self._print, key_sequence="Ctrl+P")
        self._file_menu.add_separator()
        self._quit_action = self._file_menu.add_menu_item(_("Exit"), self._request_exit, key_sequence="quit", role="quit")

        self._undo_action = self._edit_menu.add_menu_item(_("Undo"), self._undo, key_sequence="undo")
        self._redo_action = self._edit_menu.add_menu_item(_("Redo"), self._redo, key_sequence="redo")
        self._edit_menu.add_separator()
        self._cut_action = self._edit_menu.add_menu_item(_("Cut"), self._cut, key_sequence="cut")
        self._copy_action = self._edit_menu.add_menu_item(_("Copy"), self._copy, key_sequence="copy")
        self._paste_action = self._edit_menu.add_menu_item(_("Paste"), self._paste, key_sequence="paste")
        self._delete_action = self._edit_menu.add_menu_item(_("Delete"), self._delete, key_sequence="delete")
        self._select_all_action = self._edit_menu.add_menu_item(_("Select All"), self._select_all, key_sequence="select-all")
        self._edit_menu.add_separator()

        self._minimize_action = self._window_menu.add_menu_item(_("Minimize"), self._minimize)
        self._zoom_action = self._window_menu.add_menu_item(_("Zoom"), self._zoom)
        self._bring_to_front_action = self._window_menu.add_menu_item(_("Bring to Front"), self._bring_to_front)

        self._file_menu.on_about_to_show = self._file_menu_about_to_show
        self._edit_menu.on_about_to_show = self._edit_menu_about_to_show
        self._window_menu.on_about_to_show = self._window_menu_about_to_show

    def _adjust_menus(self) -> None:
        # called when key may be shortcut
        self._file_menu_about_to_show()
        self._edit_menu_about_to_show()
        self._window_menu_about_to_show()

    def _request_exit(self) -> None:
        if self.app:
            self.app.exit()

    def request_close(self) -> None:
        self.__document_window.request_close()

    def _register_ui_activity(self) -> None:
        pass

    def finish_periodic(self) -> None:
        # recognize when we're running as test and finish out periodic operations
        if not self.__document_window.has_event_loop:
            self.periodic()

    def periodic(self) -> None:
        self.__periodic_queue.perform_tasks()
        self.__periodic_set.perform_tasks()
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        if self.app:
            self.app.periodic()

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop

    def attach_widget(self, widget):
        self.__document_window.attach(widget)

    def detach_widget(self):
        self.__document_window.detach()

    def about_to_show(self) -> None:
        if self.__persistent_id:
            geometry = self.ui.get_persistent_string("{}/Geometry".format(self.__persistent_id))
            state = self.ui.get_persistent_string("{}/State".format(self.__persistent_id))
            self.restore(geometry, state)
        self.__shown = True

    def about_to_close(self, geometry: str, state: str) -> None:
        # this method is invoked when the low level window is about to close.
        # subclasses can override this method to save geometry and state.
        if callable(self.on_close):
            self.on_close()
        # this call will close this object and subsequently the ui window, but not the low level window itself.
        # it will, however, delete the widget hierarchy as a consequence of closing the ui window.
        # so care must be taken to not close windows in cases where the widget triggering the close is still in use.
        self.close()

    def refocus_widget(self, widget):
        widget.refocus()

    def __save_bounds(self):
        if self.__shown and self.__persistent_id:
            geometry, state = self.save()
            self.ui.set_persistent_string("{}/Geometry".format(self.__persistent_id), geometry)
            self.ui.set_persistent_string("{}/State".format(self.__persistent_id), state)

    def activation_changed(self, activated: bool) -> None:
        pass

    def size_changed(self, width: int, height: int) -> None:
        self.__save_bounds()

    def position_changed(self, x: int, y: int) -> None:
        self.__save_bounds()

    def key_pressed(self, key: UserInterface.Key) -> bool:
        return False

    def key_released(self, key: UserInterface.Key) -> bool:
        if key.modifiers.control and key.key:
            self._adjust_menus()
        return False

    def drag(self, mime_data: UserInterface.MimeData, thumbnail, hot_spot_x, hot_spot_y) -> None:
        self.__document_window.root_widget.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)

    @property
    def title(self) -> str:
        return self.__document_window.title

    @title.setter
    def title(self, value: str) -> None:
        self.__document_window.title = value

    def get_file_paths_dialog(self, title: str, directory: str, filter: str, selected_filter: str=None) -> (typing.List[str], str, str):
        return self.__document_window.get_file_paths_dialog(title, directory, filter, selected_filter)

    def get_file_path_dialog(self, title, directory, filter, selected_filter=None):
        return self.__document_window.get_file_path_dialog(title, directory, filter, selected_filter)

    def get_save_file_path(self, title, directory, filter, selected_filter=None):
        return self.__document_window.get_save_file_path(title, directory, filter, selected_filter)

    def create_dock_widget(self, widget, panel_id, title, positions, position):
        return self.__document_window.create_dock_widget(widget, panel_id, title, positions, position)

    def tabify_dock_widgets(self, dock_widget1, dock_widget2):
        return self.__document_window.tabify_dock_widgets(dock_widget1, dock_widget2)

    @property
    def screen_size(self):
        return self.__document_window.screen_size

    @property
    def screen_logical_dpi(self):
        return self.__document_window.screen_logical_dpi

    @property
    def screen_physical_dpi(self):
        return self.__document_window.screen_physical_dpi

    @property
    def display_scaling(self):
        return self.__document_window.display_scaling

    def get_font_metrics(self, font, text):
        return self.ui.get_font_metrics(font, text)

    @property
    def focus_widget(self):
        focus_widget = self.__document_window.focus_widget
        if focus_widget:
            return focus_widget
        for dock_widget in self.dock_widgets:
            focus_widget = dock_widget.focus_widget
            if focus_widget:
                return focus_widget
        return None

    @property
    def dock_widgets(self):
        return self.__document_window.dock_widgets

    def show(self, *, size: Geometry.IntSize=None, position: Geometry.IntPoint=None) -> None:
        self.__document_window.show(size=size, position=position)

    def add_menu(self, title: str):
        return self.__document_window.add_menu(title)

    def insert_menu(self, title: str, before_menu):
        return self.__document_window.insert_menu(title, before_menu)

    def create_sub_menu(self):
        return self.ui.create_sub_menu(self.__document_window)

    def create_context_menu(self):
        return self.ui.create_context_menu(self.__document_window)

    def restore(self, geometry: str, state: str) -> None:
        self.__document_window.restore(geometry, state)

    def save(self) -> (str, str):
        return self.__document_window.save()

    # tasks can be added in two ways, queued or added
    # queued tasks are guaranteed to be executed in the order queued.
    # added tasks are only executed if not replaced before execution.
    # added tasks do not guarantee execution order or execution at all.

    def add_task(self, key, task):
        assert task
        self.__periodic_set.add_task(key + str(id(self)), task)

    def clear_task(self, key):
        self.__periodic_set.clear_task(key + str(id(self)))

    def queue_task(self, task):
        assert task
        self.__periodic_queue.put(task)

    def clear_queued_tasks(self):
        self.__periodic_queue.clear_tasks()

    def handle_quit(self):
        self.app.exit()

    def _dispatch_any_to_focus_widget(self, method: str, *args, **kwargs) -> bool:
        focus_widget = self.focus_widget
        if focus_widget and focus_widget._dispatch_any(method, *args, **kwargs):
                return True
        if hasattr(self, method) and getattr(self, method)(*args, **kwargs):
                return True
        return False

    def _get_menu_item_state(self, command_id: str) -> typing.Optional[UserInterface.MenuItemState]:
        # if there is a specific menu item state for the command_id, use it
        # otherwise, if the handle method exists, return an enabled menu item
        # otherwise, don't handle
        handle_method = "handle_" + command_id
        menu_item_state_method = "get_" + command_id + "_menu_item_state"
        if hasattr(self, menu_item_state_method):
            menu_item_state = getattr(self, menu_item_state_method)()
            if menu_item_state:
                return menu_item_state
        if hasattr(self, handle_method):
            return UserInterface.MenuItemState(title=None, enabled=True, checked=False)
        return None

    def _get_focus_widget_menu_item_state(self, command_id: str) -> typing.Optional[UserInterface.MenuItemState]:
        focus_widget = self.focus_widget
        if focus_widget:
            menu_item_state = focus_widget._get_menu_item_state(command_id)
            if menu_item_state:
                return menu_item_state
        return self._get_menu_item_state(command_id)

    # standard menu items

    def _file_menu_about_to_show(self):
        self._close_action.enabled = True
        self._page_setup_action.apply_state(self._get_focus_widget_menu_item_state("page_setup"))
        self._print_action.apply_state(self._get_focus_widget_menu_item_state("print"))
        self._quit_action.enabled = True

    def _edit_menu_about_to_show(self):
        self._undo_action.apply_state(self._get_focus_widget_menu_item_state("undo"))
        self._redo_action.apply_state(self._get_focus_widget_menu_item_state("redo"))
        self._cut_action.apply_state(self._get_focus_widget_menu_item_state("cut"))
        self._copy_action.apply_state(self._get_focus_widget_menu_item_state("copy"))
        self._paste_action.apply_state(self._get_focus_widget_menu_item_state("paste"))
        self._delete_action.apply_state(self._get_focus_widget_menu_item_state("delete"))
        self._select_all_action.apply_state(self._get_focus_widget_menu_item_state("select_all"))

    def _window_menu_about_to_show(self):
        self._minimize_action.apply_state(self._get_focus_widget_menu_item_state("minimize"))
        self._zoom_action.apply_state(self._get_focus_widget_menu_item_state("zoom"))
        self._bring_to_front_action.apply_state(self._get_focus_widget_menu_item_state("bring_to_front"))

    def _page_setup(self):
        self._dispatch_any_to_focus_widget("handle_page_setup")

    def _print(self):
        self._dispatch_any_to_focus_widget("handle_print")

    def _cut(self):
        self._dispatch_any_to_focus_widget("handle_cut")

    def _copy(self):
        self._dispatch_any_to_focus_widget("handle_copy")

    def _paste(self):
        self._dispatch_any_to_focus_widget("handle_paste")

    def _delete(self):
        self._dispatch_any_to_focus_widget("handle_delete")

    def _select_all(self):
        self._dispatch_any_to_focus_widget("handle_select_all")

    def _undo(self):
        self._dispatch_any_to_focus_widget("handle_undo")

    def _redo(self):
        self._dispatch_any_to_focus_widget("handle_redo")

    def _minimize(self):
        self._dispatch_any_to_focus_widget("handle_minimize")

    def _zoom(self):
        self._dispatch_any_to_focus_widget("handle_zoom")

    def _bring_to_front(self):
        self._dispatch_any_to_focus_widget("bring_to_front")



"""
Preference dialog.
"""

# standard libraries
import gettext

# types
from typing import AbstractSet

# third party libraries
# None

# local libraries
from nion.ui import Dialog
from nion.ui import Widgets
from nion.utils import Event
from nion.utils import Selection

_ = gettext.gettext


class Singleton(type):
    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class PreferencesManager(metaclass=Singleton):
    def __init__(self):
        self.preference_pane_delegates = list()
        self.preference_pane_delegates_changed_event = Event.Event()

    def register_preference_pane(self, preference_pane_delegate):
        assert not preference_pane_delegate in self.preference_pane_delegates
        self.preference_pane_delegates.append(preference_pane_delegate)
        self.preference_pane_delegates_changed_event.fire()

    def unregister_preference_pane(self, preference_pane_delegate):
        assert preference_pane_delegate in self.preference_pane_delegates
        self.preference_pane_delegates.remove(preference_pane_delegate)
        self.preference_pane_delegates_changed_event.fire()


class PreferencesDialog(Dialog.ActionDialog):
    def __init__(self, ui, app):
        super().__init__(ui, _("Preferences"), app)

        self.ui = ui
        self.document_controller = self

        self._create_menus()

        properties = dict()
        properties["min-height"] = 400
        properties["min-width"] = 800

        preference_pane_delegates = list()
        preference_pane_delegate_id_ref = [None]

        content_stack = ui.create_stack_widget()

        def change_selection(indexes):
            index = list(indexes)[0]
            assert 0 <= index < len(preference_pane_delegates)
            content_stack.current_index = index
            preference_pane_delegate_id_ref[0] = preference_pane_delegates[index].identifier

        selector_list_widget = Widgets.StringListWidget(ui, selection_style=Selection.Style.single_or_none)
        selector_list_widget.on_selection_changed = change_selection

        row = self.ui.create_row_widget(properties={"min-width": 640, "min-height": 320})
        selector_column = self.ui.create_column_widget(properties={"width": 200})
        selector_row = ui.create_row_widget()
        selector_row.add_spacing(8)
        selector_row.add(selector_list_widget)
        selector_row.add_spacing(8)
        selector_column.add_spacing(8)
        selector_column.add(selector_row)
        selector_column.add_spacing(8)
        content_column = self.ui.create_column_widget()
        content_column.add(content_stack)
        row.add(selector_column)
        row.add(content_column)
        self.content.add(row)

        self.add_button(_("Done"), lambda: True)

        def rebuild():
            content_stack.remove_all()
            preference_pane_delegates.clear()
            preference_pane_delegate_id = preference_pane_delegate_id_ref[0]
            items = list()
            selected_index = 0
            for index, preference_pane_delegate in enumerate(PreferencesManager().preference_pane_delegates):
                preference_pane_delegates.append(preference_pane_delegate)
                content_column_widget = ui.create_column_widget()
                content_column_widget.add_spacing(12)
                content_column_widget.add(preference_pane_delegate.build(ui, event_loop=self.event_loop))
                content_column_widget.add_spacing(12)
                content_row_widget = ui.create_row_widget()
                content_row_widget.add_spacing(12)
                content_row_widget.add(content_column_widget)
                content_row_widget.add_spacing(12)
                content_stack.add(content_row_widget)
                items.append(preference_pane_delegate.label)
                if preference_pane_delegate.identifier == preference_pane_delegate_id:
                    selected_index = index
            change_selection({selected_index})
            selector_list_widget.items = items

        self.__preference_pane_delegates_changed_listener = PreferencesManager().preference_pane_delegates_changed_event.listen(rebuild)

        rebuild()

    def close(self):
        self.__preference_pane_delegates_changed_listener.close()
        self.__preference_pane_delegates_changed_listener = None
        super().close()



"""
    List of drawable cells.

    TODO: ListCanvasItem should handle dragging multiple items
    TODO: ListCanvasItem should handle keyboard navigation
    TODO: ListCanvasItem should allow drag selection to select multiple
"""

# futures
from __future__ import absolute_import
from __future__ import division

# standard libraries
# none

# third party libraries
# none

# local libraries
from . import CanvasItem
from nion.utils import Geometry


class ListCanvasItem(CanvasItem.AbstractCanvasItem):
    """
    Takes a delegate that supports the following properties, methods, and optional methods:

    Properties:
        item_count: the number of items to be displayed
        items: the items to be displayed

    Methods:
        paint_item(drawing_context, index, rect, is_selected): paint the cell for index at the position

    Optional methods:
        content_menu_event(index, x, y, gx, gy): called when user wants context menu for given index
        key_pressed(key): called when user presses a key
        delete_pressed(): called when user presses delete key
        drag_started(index, x, y, modifiers): called when user begins drag with given index
    """

    def __init__(self, delegate, selection, item_height=80):
        super().__init__()
        # store parameters
        self.__delegate = delegate
        self.__selection = selection
        self.__selection_changed_listener = self.__selection.changed_event.listen(self.update)
        # configure super
        self.wants_mouse_events = True
        self.focusable = True
        # internal variables
        self.__mouse_pressed = False
        self.__mouse_pressed_for_dragging = False
        self.__mouse_index = None
        self.__mouse_position = None
        self.__mouse_dragging = False
        self.__dropping = True
        self.__drop_before_index = None
        self.__drop_index = None
        self.__item_height = item_height

    def close(self):
        self.__selection_changed_listener.close()
        self.__selection_changed_listener = None
        super().close()

    def detach_delegate(self):
        self.__delegate = None

    def update_layout(self, canvas_origin, canvas_size, *, immediate=False):
        """Override from abstract canvas item.

        Adjust the canvas height based on the constraints.
        """
        canvas_size = Geometry.IntSize.make(canvas_size)
        canvas_size = Geometry.IntSize(height=self.__calculate_layout_height(), width=canvas_size.width)
        super().update_layout(canvas_origin, canvas_size, immediate=immediate)

    def wheel_changed(self, x, y, dx, dy, is_horizontal):
        dy = dy if not is_horizontal else 0.0
        new_canvas_origin = Geometry.IntPoint.make(self.canvas_origin) + Geometry.IntPoint(x=0, y=dy)
        self.update_layout(new_canvas_origin, self.canvas_size)
        self.update()
        return True

    def handle_tool_tip(self, x: int, y: int, gx: int, gy: int) -> bool:
        max_index = self.__delegate.item_count
        mouse_index = y // self.__item_height
        if mouse_index >= 0 and mouse_index < max_index:
            if self.__delegate and hasattr(self.__delegate, "item_tool_tip"):
                text = self.__delegate.item_tool_tip(mouse_index)
                if text:
                    self.show_tool_tip_text(text, gx, gy)
                    return True
        return super().handle_tool_tip(x, y, gx, gy)

    def __calculate_layout_height(self):
        item_count = self.__delegate.item_count if self.__delegate else 0
        return item_count * self.__item_height

    def __rect_for_index(self, index: int) -> Geometry.IntRect:
        canvas_bounds = self.canvas_bounds
        item_width = int(canvas_bounds.width)
        item_height = self.__item_height
        return Geometry.IntRect(origin=Geometry.IntPoint(y=index * item_height, x=0),
                                size=Geometry.IntSize(width=item_width, height=item_height))

    def _repaint_visible(self, drawing_context, visible_rect):
        if self.__delegate:
            canvas_bounds = self.canvas_bounds

            item_width = int(canvas_bounds.width)
            item_height = self.__item_height

            with drawing_context.saver():
                items = self.__delegate.items
                max_index = len(items)
                top_visible_row = visible_rect.top // item_height
                bottom_visible_row = visible_rect.bottom // item_height
                for index in range(top_visible_row, bottom_visible_row + 1):
                    if 0 <= index < max_index:
                        rect = Geometry.IntRect(origin=Geometry.IntPoint(y=index * item_height, x=0),
                                                size=Geometry.IntSize(width=item_width, height=item_height))
                        if rect.intersects_rect(visible_rect):
                            is_selected = self.__selection.contains(index)
                            if is_selected:
                                with drawing_context.saver():
                                    drawing_context.begin_path()
                                    drawing_context.rect(rect.left, rect.top, rect.width, rect.height)
                                    drawing_context.fill_style = "#3875D6" if self.focused else "#DDD"
                                    drawing_context.fill()
                            self.__delegate.paint_item(drawing_context, items[index], rect, is_selected)
                            if index == self.__drop_index:
                                with drawing_context.saver():
                                    drop_border_width = 2.5
                                    rect_in = rect.inset(drop_border_width / 2, drop_border_width / 2)
                                    drawing_context.begin_path()
                                    drawing_context.rect(rect_in.left, rect_in.top, rect_in.width, rect_in.height)
                                    drawing_context.line_width = drop_border_width
                                    drawing_context.stroke_style = "rgba(56, 117, 214, 0.8)"
                                    drawing_context.stroke()

    def _repaint(self, drawing_context):
        self._repaint_visible(drawing_context, self.canvas_bounds)

    def context_menu_event(self, x, y, gx, gy):
        if self.__delegate:
            max_index = self.__delegate.item_count
            mouse_index = y // self.__item_height
            if mouse_index >= 0 and mouse_index < max_index:
                if not self.__selection.contains(mouse_index):
                    self.__selection.set(mouse_index)
                if self.__delegate and hasattr(self.__delegate, "context_menu_event") and self.__delegate.context_menu_event:
                    return self.__delegate.context_menu_event(mouse_index, x, y, gx, gy)
                # TODO: delete soon. only here for backwards compatibility.
                if self.__delegate and hasattr(self.__delegate, "on_context_menu_event") and self.__delegate.on_context_menu_event:
                    return self.__delegate.on_context_menu_event(mouse_index, x, y, gx, gy)
            else:
                if self.__delegate and hasattr(self.__delegate, "context_menu_event") and self.__delegate.context_menu_event:
                    return self.__delegate.context_menu_event(None, x, y, gx, gy)
                # TODO: delete soon. only here for backwards compatibility.
                if self.__delegate and hasattr(self.__delegate, "on_context_menu_event") and self.__delegate.on_context_menu_event:
                    return self.__delegate.on_context_menu_event(None, x, y, gx, gy)
        return False

    def mouse_double_clicked(self, x, y, modifiers):
        max_index = self.__delegate.item_count
        mouse_index = y // self.__item_height
        if mouse_index >= 0 and mouse_index < max_index:
            if not self.__selection.contains(mouse_index):
                self.__selection.set(mouse_index)
            if self.__delegate and hasattr(self.__delegate, "item_selected") and self.__delegate.item_selected:
                return self.__delegate.item_selected(mouse_index)
        return super().mouse_double_clicked(x, y, modifiers)

    def mouse_pressed(self, x, y, modifiers):
        if self.__delegate:
            mouse_index = y // self.__item_height
            max_index = self.__delegate.item_count
            if mouse_index >= 0 and mouse_index < max_index:
                self.__mouse_index = mouse_index
                self.__mouse_pressed = True
                handled = False
                if self.__delegate and hasattr(self.__delegate, "mouse_pressed_in_item") and self.__delegate.mouse_pressed_in_item:
                    handled = self.__delegate.mouse_pressed_in_item(mouse_index, Geometry.IntPoint(y=y - mouse_index * self.__item_height, x=x), modifiers)
                    if handled:
                        self.__mouse_index = None  # prevent selection handling
                if not handled and not modifiers.shift and not modifiers.control:
                    self.__mouse_pressed_for_dragging = True
                    self.__mouse_position = Geometry.IntPoint(y=y, x=x)
                return True
        return super().mouse_pressed(x, y, modifiers)

    def __mouse_released(self, x, y, modifiers, do_select):
        if self.__delegate and self.__mouse_pressed and do_select:
            # double check whether mouse_released has been called explicitly as part of a drag.
            # see https://bugreports.qt.io/browse/QTBUG-40733
            mouse_index = self.__mouse_index
            max_index = self.__delegate.item_count
            if mouse_index is not None and mouse_index >= 0 and mouse_index < max_index:
                if modifiers.shift:
                    self.__selection.extend(mouse_index)
                elif modifiers.control:
                    self.__selection.toggle(mouse_index)
                else:
                    self.__selection.set(mouse_index)
        self.__mouse_pressed = False
        self.__mouse_pressed_for_dragging = False
        self.__mouse_index = None
        self.__mouse_position = None
        self.__mouse_dragging = False
        return True

    def mouse_released(self, x, y, modifiers):
        return self.__mouse_released(x, y, modifiers, True)

    def mouse_position_changed(self, x, y, modifiers):
        if self.__mouse_pressed_for_dragging:
            if not self.__mouse_dragging and Geometry.distance(self.__mouse_position, Geometry.IntPoint(y=y, x=x)) > 8:
                self.__mouse_dragging = True
                if self.__delegate and hasattr(self.__delegate, "drag_started") and self.__delegate.drag_started:
                    root_container = self.root_container
                    if root_container:
                        root_container.bypass_request_focus()
                    self.__delegate.drag_started(self.__mouse_index, x, y, modifiers)
                    # once a drag starts, mouse release will not be called; call it here instead
                    self.__mouse_released(x, y, modifiers, False)
                # TODO: delete soon. only here for backwards compatibility.
                if self.__delegate and hasattr(self.__delegate, "on_drag_started") and self.__delegate.on_drag_started:
                    root_container = self.root_container
                    if root_container:
                        root_container.bypass_request_focus()
                    self.__delegate.on_drag_started(self.__mouse_index, x, y, modifiers)
                    # once a drag starts, mouse release will not be called; call it here instead
                    self.__mouse_released(x, y, modifiers, False)
                return True
        return super().mouse_position_changed(x, y, modifiers)

    def __make_selection_visible(self, top):
        if self.__delegate:
            selected_indexes = list(self.__selection.indexes)
            if len(selected_indexes) > 0 and self.canvas_bounds is not None:
                min_index = min(selected_indexes)
                max_index = max(selected_indexes)
                min_rect = self.__rect_for_index(min_index)
                max_rect = self.__rect_for_index(max_index)
                visible_rect = getattr(self.container, "visible_rect", None)
                if visible_rect is not None:
                    if top:
                        if min_rect.top < visible_rect.top:
                            self.update_layout(Geometry.IntPoint(y=-min_rect.top, x=self.canvas_origin.x), self.canvas_size)
                        elif min_rect.bottom > visible_rect.bottom:
                            self.update_layout(Geometry.IntPoint(y=-min_rect.bottom + visible_rect.height, x=self.canvas_origin.x), self.canvas_size)
                    else:
                        if max_rect.bottom > visible_rect.bottom:
                            self.update_layout(Geometry.IntPoint(y=-max_rect.bottom + visible_rect.height, x=self.canvas_origin.x), self.canvas_size)
                        elif max_rect.top < visible_rect.top:
                            self.update_layout(Geometry.IntPoint(y=-max_rect.top, x=self.canvas_origin.x), self.canvas_size)

    def make_selection_visible(self):
        self.__make_selection_visible(True)

    def key_pressed(self, key):
        if self.__delegate:
            if self.__delegate and hasattr(self.__delegate, "key_pressed") and self.__delegate.key_pressed:
                if self.__delegate.key_pressed(key):
                    return True
            # TODO: delete soon. only here for backwards compatibility.
            if self.__delegate and hasattr(self.__delegate, "on_key_pressed") and self.__delegate.on_key_pressed:
                if self.__delegate.on_key_pressed(key):
                    return True
            if key.is_delete:
                return self.handle_delete()
            if key.is_enter_or_return:
                if self.__delegate and hasattr(self.__delegate, "item_selected") and self.__delegate.item_selected:
                    indexes = self.__selection.indexes
                    if len(indexes) == 1:
                        return self.__delegate.item_selected(list(indexes)[0])
            if key.is_up_arrow:
                new_index = None
                indexes = self.__selection.indexes
                if len(indexes) > 0:
                    new_index = max(min(indexes) - 1, 0)
                elif self.__delegate.item_count > 0:
                    new_index = self.__delegate.item_count - 1
                if new_index is not None:
                    if key.modifiers.shift:
                        self.__selection.extend(new_index)
                    else:
                        self.__selection.set(new_index)
                self.__make_selection_visible(top=True)
                return True
            if key.is_down_arrow:
                new_index = None
                indexes = self.__selection.indexes
                if len(indexes) > 0:
                    new_index = min(max(indexes) + 1, self.__delegate.item_count - 1)
                elif self.__delegate.item_count > 0:
                    new_index = 0
                if new_index is not None:
                    if key.modifiers.shift:
                        self.__selection.extend(new_index)
                    else:
                        self.__selection.set(new_index)
                self.__make_selection_visible(top=False)
                return True
        return super().key_pressed(key)

    def drag_enter(self, mime_data):
        self.__dropping = True
        self.update()
        return "ignore"

    def drag_move(self, mime_data, x, y):
        mouse_index = y // self.__item_height
        max_index = self.__delegate.item_count
        drop_index = None
        if mouse_index >= 0 and mouse_index < max_index:
            drop_index = mouse_index
            if self.__delegate and hasattr(self.__delegate, "item_can_drop_mime_data") and callable(self.__delegate.item_can_drop_mime_data):
                if not self.__delegate.item_can_drop_mime_data(mime_data, "move", drop_index):
                    drop_index = None
            else:
                drop_index = None
        if drop_index != self.__drop_index:
            self.__drop_index = drop_index
            self.update()

    def drag_leave(self):
        self.__dropping = False
        self.__drop_index = None
        self.update()
        return False

    def drop(self, mime_data, x, y):
        drop_index = self.__drop_index
        self.__dropping = False
        self.__drop_index = None
        self.update()
        if drop_index is not None:
            if self.__delegate and hasattr(self.__delegate, "item_drop_mime_data") and callable(self.__delegate.item_drop_mime_data):
                return self.__delegate.item_drop_mime_data(mime_data, "move", drop_index)
        return "ignore"

    def handle_select_all(self):
        if self.__delegate:
            self.__selection.set_multiple(set(range(self.__delegate.item_count)))
            return True
        return False

    def handle_delete(self):
        if self.__delegate and hasattr(self.__delegate, "delete_pressed") and self.__delegate.delete_pressed:
            self.__delegate.delete_pressed()
        # TODO: delete soon. only here for backwards compatibility.
        if self.__delegate and hasattr(self.__delegate, "on_delete_pressed") and self.__delegate.on_delete_pressed:
            self.__delegate.on_delete_pressed()
        return True

    def size_to_content(self) -> None:
        """Size the canvas item to the height of the items."""
        new_sizing = self.copy_sizing()
        new_sizing.minimum_height = self.__item_height * self.__delegate.item_count
        new_sizing.maximum_height = self.__item_height * self.__delegate.item_count
        self.update_sizing(new_sizing)



from nion.utils import Model


class Handler:

    def __init__(self):
        self.slider_value_model = Model.PropertyModel(50)
        self.vv = 25
        self.slider2 = None

    def reset(self, widget):
        self.slider_value_model.value = 50
        self.slider2.value = 50

    def value_changed(self, widget, value):
        print(f"V {value}")

    def slider_moved(self, widget, value):
        print(f"M {value}")


def construct_ui(ui):

    slider = ui.create_slider(value="@binding(slider_value_model.value)")

    slider2 = ui.create_slider(name="slider2", value="vv", on_value_changed="value_changed", on_slider_moved="slider_moved")

    label = ui.create_label(text="@binding(slider_value_model.value)")

    button = ui.create_push_button(text="Reset to 50", on_clicked="reset")

    return ui.create_column(slider, slider2, label, button, spacing=8)



from nion.utils import Converter
from nion.utils import Model


class Handler:
    hello3_converter = Converter.FloatToStringConverter(format="{0:#.4g}")
    hello3_model = Model.PropertyModel(5.6)


def construct_ui(ui):
    hello3_line_edit = ui.create_line_edit(placeholder_text="Enter Text", text="@binding(hello3_model.value, converter=hello3_converter)")
    hello3_label = ui.create_label(text="@binding(hello3_model.value, converter=hello3_converter)")
    hello3 = ui.create_column(hello3_label, hello3_line_edit, spacing=8)
    return hello3



from nion.utils import Model


class Handler:

    slider_value_model = Model.PropertyModel(50)

    def reset(self, widget):
        self.slider_value_model.value = 50


def construct_ui(ui):

    slider = ui.create_slider(value="@binding(slider_value_model.value)")

    progress_bar = ui.create_progress_bar(value="@binding(slider_value_model.value)")

    label = ui.create_label(text="@binding(slider_value_model.value)")

    button = ui.create_push_button(text="Reset to 50", on_clicked="reset")

    group_column = ui.create_column(slider, label, spacing=8)

    group = ui.create_group(group_column, title="Group", margin=4)

    return ui.create_column(progress_bar, group, button, spacing=8)



from nion.ui import Declarative

class Handler:

    width = "20"
    height = "30"

    width_field = None
    height_field = None

    def width_changed(self, widget, value):
        self.width = value
        print(f"New width {self.width}")

    def height_changed(self, widget, value):
        self.height = value
        print(f"New height {self.height}")

    def reset(self, widget):
        self.width_field.handler.value_widget.text = "20"
        self.height_field.handler.value_widget.text = "30"

    def create_handler(self, component_id: str, **kwargs):

        class FieldHandler:

            def __init__(self):
                self.label_widget = None
                self.value_widget = None
                self.on_value_changed = None
                self.label = None
                self.value = None

            def init_handler(self):
                # when this is called, all fields will be populated
                self.label_widget.text = self.label
                self.value_widget.text = self.value

            def value_changed(self, widget, text):
                self.value = text
                if widget.focused:
                    widget.select_all()
                if callable(self.on_value_changed):
                    self.on_value_changed(value=text)

        if component_id == "field":
            return FieldHandler()

    @property
    def resources(self):
        ui = Declarative.DeclarativeUI()
        field_label = ui.create_label(name="label_widget")
        field_line_edit = ui.create_line_edit(name="value_widget", on_editing_finished="value_changed")
        field = ui.create_row(field_label, field_line_edit, ui.create_stretch(), spacing=8)
        field_events = [{"event": "on_value_changed", "parameters": ["value"]}]
        field_component = ui.define_component(content=field, component_id="field", events=field_events)
        return {"field": field_component}


def construct_ui(ui):
    field_width = ui.create_component_instance("field", {"label": "Width", "value": "20"}, name="width_field", on_value_changed="width_changed")
    field_height = ui.create_component_instance("field", {"label": "Height", "value": "30"}, name="height_field", on_value_changed="height_changed")
    reset_button = ui.create_push_button(text="Reset", on_clicked="reset")
    return ui.create_column(field_width, field_height, reset_button, spacing=8)



"""
A basic class to serve as the basis of a typical one window application.
"""
# standard libraries
import logging

# local libraries
# None


class LoggingHandler(logging.StreamHandler):

    def __init__(self):
        super().__init__()
        self.__records = list()

    def emit(self, record):
        super().emit(record)
        if self.__records is not None:
            self.__records.append(record)

    def take_records(self):
        records = self.__records
        self.__records = None
        return records


logging_handler = LoggingHandler()


class Application:
    """A basic application class.

    Subclass this class and implement the start method. The start method should create a document window that will be
    the focus of the UI.

    Pass the desired user interface to the init method. Then call initialize and start.
    """

    def __init__(self, ui, *, on_start=None):
        self.ui = ui
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        logger.addHandler(logging_handler)
        self.window = None
        self.on_start = on_start

    def initialize(self):
        """Initialize. Separate from __init__ so that overridden methods can be called."""
        pass

    def deinitialize(self):
        self.ui.close()

    def run(self):
        """Alternate start which allows ui to control event loop."""
        self.ui.run(self)

    def start(self):
        """The start method should create a window that will be the focus of the UI."""
        if self.on_start:
            return self.on_start()
        raise NotImplemented()

    def exit(self):
        """The exit method should request to close or close the window."""
        raise NotImplemented()

    def periodic(self):
        """The periodic method can be overridden to implement periodic behavior."""
        pass


def make_ui(bootstrap_args):
    if "proxy" in bootstrap_args:
        from nion.ui import QtUserInterface
        proxy = bootstrap_args["proxy"]
        return QtUserInterface.QtUserInterface(proxy)
    elif "pyqt" in bootstrap_args:
        from nion.ui import QtUserInterface
        from nion.ui import PyQtProxy
        return QtUserInterface.QtUserInterface(PyQtProxy.PyQtProxy())
    elif "server" in bootstrap_args:
        from nion.ui import CanvasUI
        server = bootstrap_args["server"]
        return CanvasUI.CanvasUserInterface(server.draw, server.get_font_metrics)
    else:
        return None



# standard libraries
import gettext
import re
import typing

# local libraries
from nion.ui import Application
from nion.ui import Dialog
from nion.ui import Window
from nion.ui import UserInterface
from nion.ui import Widgets
from nion.utils import Binding
from nion.utils import Registry


UIDescription = typing.Dict  # when napolean works: typing.NewType("UIDescription", typing.Dict)
UIResources = typing.Dict  # when napolean works: typing.NewType("UIResources", typing.Dict)
UIPoints = int  # when napolean works: typing.NewType("UIPoints", int)
UILabel = str
UIIdentifier = str  # typing.NewType("UIIndentifier", str)
UICallableIdentifier = str  # typing.NewType("UICallableIdentifier", str)
UIWidget = UserInterface.Widget
UIKey = UserInterface.Key

_ = gettext.gettext


class DeclarativeUI:

    # ----: row
    # ----: column
    # ----: spacing
    # ----: stretch
    # ----: stack
    # ----: tab
    # ----: label
    # ----: text edit
    # ----: line edit
    # TODO: scroll area
    # ----: group box
    # ----: status bar
    # TODO: tool tips
    # TODO: expander
    # TODO: border
    # ----: push button
    # ----: check box
    # ----: combo box
    # ----: radio buttons
    # TODO: splitter
    # TODO: image
    # ----: component
    # TODO: part
    # TODO: data view
    # TODO: list view
    # TODO: tree view
    # ----: slider
    # TODO: menus
    # TODO: context menu
    # ----: progress bar
    # TODO: key handler
    # TODO: key validator
    # TODO: field validator
    # TODO: canvas
    # TODO: dock panels
    # TODO: windows
    # TODO: thumbnails
    # TODO: display panels
    # TODO: periodic
    # TODO: focus handler
    # ----: bindings
    # TODO: commands
    # TODO: standard dialog boxes, open, save, print, confirm
    # TODO: all static text (checkbox 'text') should be bindable
    # TODO: how to define resources for a sub component?
    # TODO: windows: fit to content; fixed sizes
    # TODO: tab label should be bindable
    # TODO: window and dialog title should be bindable
    # TODO: placeholder text should be bindable

    def __init__(self):
        pass

    def __process_common_properties(self, d: typing.MutableMapping, **kwargs) -> None:
        common_properties = (
            "enabled",
            "visible",
            "width",
            "min_width",
            "max_width",
            "height",
            "min_height",
            "max_height",
            "tool_tip"
        )
        for k in common_properties:
            if k in kwargs and kwargs[k] is not None:
                d[k] = kwargs[k]

    def create_column(self, *children: UIDescription, items: UIIdentifier=None, item_component_id: str=None, spacing: UIPoints=None, margin: UIPoints=None, **kwargs) -> UIDescription:
        """Create a column UI description with children or dynamic items, spacing, and margin.

        The children can be passed as parameters or constructed from an observable list specified by `items` and
        `item_component_id`.

        If the children are constructed from an observable list, a component instance will be created for each item
        in the list. Adding or removing items from the list will dynamically update the children. The associated
        handler must implement the `resources` property with a value for the key `item_component_id` describing the
        component UI. It must also implement `create_handler` to create a handler for each component. The
        `create_handler` call will receive two extra keyword arguments `container` and `item` in additional to the
        `component_id`.

        Args:
            children: children to put into the column

        Keyword Args:
            items: handler observable list property from which to build child components
            item_component_id: resource identifier of component ui description for child components
            spacing: spacing between items, in points
            margin: margin, in points

        Returns:
            a UI description of the column
        """
        d = {"type": "column"}
        if items:
            d["items"] = items
        if item_component_id:
            d["item_component_id"] = item_component_id
        if spacing is not None:
            d["spacing"] = spacing
        if margin is not None:
            d["margin"] = margin
        if len(children) > 0:
            d_children = d.setdefault("children", list())
            for child in children:
                d_children.append(child)
        self.__process_common_properties(d, **kwargs)
        return d

    def create_row(self, *children: UIDescription, items: UIIdentifier=None, item_component_id: str=None, spacing: UIPoints=None, margin: UIPoints=None, **kwargs) -> UIDescription:
        """Create a row UI description with children or dynamic items, spacing, and margin.

        The children can be passed as parameters or constructed from an observable list specified by `items` and
        `item_component_id`.

        If the children are constructed from an observable list, a component instance will be created for each item
        in the list. Adding or removing items from the list will dynamically update the children. The associated
        handler must implement the `resources` property with a value for the key `item_component_id` describing the
        component UI. It must also implement `create_handler` to create a handler for each component. The
        `create_handler` call will receive two extra keyword arguments `container` and `item` in additional to the
        `component_id`.

        Args:
            children: children to put into the row

        Keyword Args:
            items: handler observable list property from which to build child components
            item_component_id: resource identifier of component ui description for child components
            spacing: spacing between items, in points
            margin: margin, in points

        Returns:
            a UI description of the row
        """
        d = {"type": "row"}
        if items:
            d["items"] = items
        if item_component_id:
            d["item_component_id"] = item_component_id
        if spacing is not None:
            d["spacing"] = spacing
        if margin is not None:
            d["margin"] = margin
        if len(children) > 0:
            d_children = d.setdefault("children", list())
            for child in children:
                d_children.append(child)
        self.__process_common_properties(d, **kwargs)
        return d

    def create_spacing(self, size: UIPoints) -> UIDescription:
        """Create a spacing UI description for a row or column.

        Keyword Args:
            size: spacing, in points

        Returns:
            a UI description of the spacing
        """
        return {"type": "spacing", "size": size}

    def create_stretch(self) -> UIDescription:
        """Create a stretch UI description for a row or column.

        Returns:
            a UI description of the stretch
        """
        return {"type": "stretch"}

    def create_tab(self, label: UILabel, content: UIDescription) -> UIDescription:
        """Create a tab UI description with a label and content.

        Args:
            label: label for the tab
            content: UI description of the content

        Returns:
            a UI description of the tab
        """
        return {"type": "tab", "label": label, "content": content}

    def create_tabs(self, *tabs: UIDescription, name: UIIdentifier=None, current_index: UIIdentifier=None, on_current_index_changed: UICallableIdentifier=None, **kwargs) -> UIDescription:
        """Create a tabs UI description with children, the current index and optional changed event.

        The children must be tabs created by :py:meth:`create_tab`.

        The current_index controls which tab is displayed.

        The on_current_index_changed callback reference takes ``widget`` and ``current_index`` parameters. The type
        signature in the handler should be ``typing.Callable[[UIWidget, int], None]``.

        Args:
            children: child tabs

        Keyword Args:
            name: handler property in which to store widget (optional)
            current_index: current index handler reference (bindable, optional)
            on_current_index_changed: callback when current index changes (optional)

        Returns:
            a UI description of the tabs
        """
        d = {"type": "tabs"}
        if len(tabs) > 0:
            d_children = d.setdefault("tabs", list())
            for child in tabs:
                d_children.append(child)
        if name is not None:
            d["name"] = name
        if current_index is not None:
            d["current_index"] = current_index
        if on_current_index_changed is not None:
            d["on_current_index_changed"] = on_current_index_changed
        self.__process_common_properties(d, **kwargs)
        return d

    def create_stack(self, *children: UIDescription, items: UIIdentifier=None, item_component_id: str=None, name: UIIdentifier=None, current_index: UIIdentifier=None, on_current_index_changed: UICallableIdentifier=None, **kwargs) -> UIDescription:
        """Create a stack UI description with children or dynamic items, the current index and optional changed event.

        The children can be passed as parameters or constructed from an observable list specified by `items` and
        `item_component_id`.

        If the children are constructed from an observable list, a component instance will be created for each item
        in the list. Adding or removing items from the list will dynamically update the children. The associated
        handler must implement the `resources` property with a value for the key `item_component_id` describing the
        component UI. It must also implement `create_handler` to create a handler for each component. The
        `create_handler` call will receive two extra keyword arguments `container` and `item` in additional to the
        `component_id`.

        The current_index controls which child is displayed.

        The on_current_index_changed callback reference takes ``widget`` and ``current_index`` parameters. The type
        signature in the handler should be ``typing.Callable[[UIWidget, int], None]``.

        Args:
            children: stack items

        Keyword Args:
            items: handler observable list property from which to build child components
            item_component_id: resource identifier of component ui description for child components
            name: handler property in which to store widget (optional)
            current_index: current index handler reference (bindable, optional)
            on_current_index_changed: callback when current index changes (optional)

        Returns:
            a UI description of the stack
        """
        d = {"type": "stack"}
        if len(children) > 0:
            d_children = d.setdefault("children", list())
            for child in children:
                d_children.append(child)
        if items:
            d["items"] = items
        if item_component_id:
            d["item_component_id"] = item_component_id
        if name is not None:
            d["name"] = name
        if current_index is not None:
            d["current_index"] = current_index
        if on_current_index_changed is not None:
            d["on_current_index_changed"] = on_current_index_changed
        self.__process_common_properties(d, **kwargs)
        return d

    def create_scroll_area(self, content: UIDescription, name: UIIdentifier=None, **kwargs) -> UIDescription:
        """Create a scroll area UI description with content and a name.

        Args:
            content: UI description of the content

        Keyword Args:
            name: handler property in which to store widget (optional)

        Returns:
            UI description of the scroll area
        """
        d = {"type": "scroll_area", "content": content}
        if name is not None:
            d["name"] = name
        self.__process_common_properties(d, **kwargs)
        return d

    def create_group(self, content: UIDescription, name: UIIdentifier=None, title: UILabel=None, margin: UIPoints=None, **kwargs) -> UIDescription:
        """Create a group UI description with content, a name, a title, and a margin.

        Args:
            content: UI description of the content

        Keyword Args:
            name: handler property in which to store widget (optional)
            title: title of the group
            margin: margin in points

        Returns:
            UI description of the group
        """
        d = {"type": "group", "content": content}
        if name is not None:
            d["name"] = name
        if title is not None:
            d["title"] = title
        if margin is not None:
            d["margin"] = margin
        self.__process_common_properties(d, **kwargs)
        return d

    def create_label(self, *, text: UILabel=None, name: UIIdentifier=None, width: int=None, min_width: int=None, **kwargs) -> UIDescription:
        """Create a label UI description with text and an optional name.

        Keyword Args:
            text: text of the label (bindable)
            name: handler property in which to store widget (optional)
            width: width in points (optional, default None)

        Returns:
            UI description of the label
        """
        d = {"type": "text_label"}
        if text is not None:
            d["text"] = text
        if name is not None:
            d["name"] = name
        self.__process_common_properties(d, **kwargs)
        return d

    def create_line_edit(self, *,
                         text: UIIdentifier=None,
                         name: UIIdentifier=None,
                         editable: bool=None,
                         placeholder_text: UILabel=None,
                         clear_button_enabled: bool=None,
                         on_editing_finished: UICallableIdentifier=None,
                         on_escape_pressed: UICallableIdentifier=None,
                         on_return_pressed: UICallableIdentifier=None,
                         on_key_pressed: UICallableIdentifier=None,
                         on_text_edited: UICallableIdentifier=None,
                         **kwargs) -> UIDescription:
        """Create a line edit UI description with text, name, placeholder, options, and events.

        The ``on_editing_finished`` callback is invoked when the user presses return or escape or when they change
        keyboard focus away from the line edit. The line edit widget and string are passed to the callback. The type
        signature in the handler should be ``typing.Callable[[UIWidget, str], None]``.

        The ``on_escape_pressed`` and ``on_return_pressed`` callbacks are invoked when the user presses escape or
        return. The line edit widget is passed and these methods must return ``True`` if they handle the key or
        ``False`` otherwise. Their type signatures in the handler should be ``typing.Callable[[UIWidget], bool]``.

        The ``on_key_pressed`` callback is invoked when the user types a key. The line edit widget and a key instance
        are passed. This method should return ``True`` if the key is handled (it will not go into the line edit field)
        and return ``False`` if not handled (it will be entered as regular text). The type signature in the handler
        should be ``typing.Callable[[UIWidget, UIKey], bool]``.

        The ``on_text_edited`` callback is invoked when the user changes the text. The line edit widget and the new text
        are passed to the callback. The type signature in the handler should be ``typing.Callable[[UIWidget, str],
        None]``.

        Keyword Args:
            text: handler reference to line edit text (bindable, required)
            name: handler property in which to store widget (optional)
            editable: whether the line edit text is editable (optional, default True)
            placeholder_text: text to display when line edit is empty (optional)
            clear_button_enabled: whether the clear button is enabled (optional, default False)
            width: width in points (optional, default None)
            on_editing_finished: callback when editing is finished (return or blur focus)
            on_escape_pressed: callback when escape is pressed, return true if handled
            on_return_pressed: callback when return is pressed, return true if handled
            on_key_pressed: callback when a key is pressed, return true if handled
            on_text_edited: callback when text is edited

        Returns:
            UI description of the line edit
        """
        d = {"type": "line_edit"}
        if text is not None:
            d["text"] = text
        if name is not None:
            d["name"] = name
        if editable is not None:
            d["editable"] = editable
        if placeholder_text is not None:
            d["placeholder_text"] = placeholder_text
        if clear_button_enabled is not None:
            d["clear_button_enabled"] = clear_button_enabled
        if on_editing_finished is not None:
            d["on_editing_finished"] = on_editing_finished
        if on_escape_pressed is not None:
            d["on_escape_pressed"] = on_escape_pressed
        if on_return_pressed is not None:
            d["on_return_pressed"] = on_return_pressed
        if on_key_pressed is not None:
            d["on_key_pressed"] = on_key_pressed
        if on_text_edited is not None:
            d["on_text_edited"] = on_text_edited
        self.__process_common_properties(d, **kwargs)
        return d

    def create_text_edit(self, *,
                         text: UIIdentifier=None,
                         name: UIIdentifier=None,
                         editable: bool=None,
                         placeholder_text: UILabel=None,
                         clear_button_enabled: bool=None,
                         on_escape_pressed: UICallableIdentifier=None,
                         on_return_pressed: UICallableIdentifier=None,
                         on_text_edited: UICallableIdentifier=None,
                         **kwargs) -> UIDescription:
        """Create a multi-line text edit UI description with text, name, placeholder, options, and events.

        The ``on_escape_pressed`` and ``on_return_pressed`` callbacks are invoked when the user presses escape or
        return. The text edit widget is passed and these methods must return ``True`` if they handle the key or
        ``False`` otherwise. Their type signatures in the handler should be ``typing.Callable[[UIWidget], bool]``.

        The ``on_text_edited`` callback is invoked when the user changes the text. The text edit widget and the new text
        are passed to the callback. The type signature in the handler should be ``typing.Callable[[UIWidget, str],
        None]``.

        Keyword Args:
            text: handler reference to line edit text (bindable, required)
            name: handler property in which to store widget (optional)
            editable: whether the line edit text is editable (optional, default True)
            placeholder_text: text to display when line edit is empty (optional)
            clear_button_enabled: whether the clear button is enabled (optional, default False)
            width: width in points (optional, default None)
            height: width in points (optional, default None)
            on_escape_pressed: callback when escape is pressed, return true if handled
            on_return_pressed: callback when return is pressed, return true if handled
            on_text_edited: callback when text is edited

        Returns:
            UI description of the text edit
        """
        d = {"type": "text_edit"}
        if text is not None:
            d["text"] = text
        if name is not None:
            d["name"] = name
        if editable is not None:
            d["editable"] = editable
        if placeholder_text is not None:
            d["placeholder_text"] = placeholder_text
        if clear_button_enabled is not None:
            d["clear_button_enabled"] = clear_button_enabled
        if on_escape_pressed is not None:
            d["on_escape_pressed"] = on_escape_pressed
        if on_return_pressed is not None:
            d["on_return_pressed"] = on_return_pressed
        if on_text_edited is not None:
            d["on_text_edited"] = on_text_edited
        self.__process_common_properties(d, **kwargs)
        return d

    def create_push_button(self, *, text: UILabel=None, icon: UIIdentifier=None, name: UIIdentifier=None, on_clicked: UICallableIdentifier=None, **kwargs) -> UIDescription:
        """Create a push button UI description with text, name, an event.

        The ``on_clicked`` callback is invoked when the user clicks the button. The widget is passed to the callback.
        The type signature in the handler should be ``typing.Callable[[UIWidget], None]``.

        Keyword Args:
            text: text of the label (bindable)
            icon: icon, bindable to numpy uint32 bgra array
            name: handler property in which to store widget (optional)
            width: width in points (optional, default None)
            on_clicked: callback when button clicked

        Returns:
            UI description of the push button
        """
        d = {"type": "push_button"}
        if text is not None:
            d["text"] = text
        if icon is not None:
            d["icon"] = icon
        if name is not None:
            d["name"] = name
        if on_clicked is not None:
            d["on_clicked"] = on_clicked
        self.__process_common_properties(d, **kwargs)
        return d

    def create_check_box(self, *,
                         text: UILabel=None,
                         name: UIIdentifier=None,
                         checked: bool=None,
                         check_state: str=None,
                         tristate: bool=None,
                         on_checked_changed: UICallableIdentifier=None,
                         on_check_state_changed: UICallableIdentifier=None,
                         **kwargs) -> UIDescription:
        """Create a check box UI description with text, name, state information, and events.

        The ``checked`` and ``check_state`` both refer to the check state. Some callers may choose to use the simpler
        ``checked`` which is a simple boolean. ``check_state`` is a string and must be one of 'checked', 'unchecked', or
        'partial'. 'partial' is only valid if ``tristate`` is ``True``.

        The ``on_checked_changed`` callback is invoked when the user changes the state of the check box. The widget and
        the new state of the check box are passed to the callback. The type signature in the handler should be
        ``typing.Callable[[UIWidget, bool], None]``.

        The ``on_check_state_changed`` callback is invoked when the user changes the state of the check box, but it also
        includes the 'partial' state if enabled. The widget and the new state of the check box are passed to the
        callback. The type signature in the handler should be ``typing.Callable[[UIWidget, str], None]``.

        Keyword Args:
            text: text of the label (bindable)
            name: handler property in which to store widget (optional)
            checked: checked state (bool)
            check_state: checked state (string: checked, unchecked, or partial)
            tristate: whether the check box is tristate or not
            on_checked_changed: callback when checked changes (optional)
            on_check_state_changed: callback when check state changes (optional)

        Returns:
            UI description of the check box
        """
        d = {"type": "check_box"}
        if text is not None:
            d["text"] = text
        if checked is not None:
            d["checked"] = checked
        if check_state is not None:
            d["check_state"] = check_state
        if tristate is not None:
            d["tristate"] = tristate
        if name is not None:
            d["name"] = name
        if on_checked_changed is not None:
            d["on_checked_changed"] = on_checked_changed
        if on_check_state_changed is not None:
            d["on_check_state_changed"] = on_check_state_changed
        self.__process_common_properties(d, **kwargs)
        return d

    def create_combo_box(self, *,
                         name: UIIdentifier=None,
                         items: typing.List[UILabel]=None,
                         items_ref: UIIdentifier=None,
                         current_index: UIIdentifier=None,
                         on_current_index_changed: UICallableIdentifier=None,
                         **kwargs):
        """Create a combo box UI description with name, items, current index, and events.

        The ``on_current_index_changed`` callback is invoked when the user changes the selected item in the combo box.
        The widget and the new index of the selected item are passed to the callback. The type signature in the handler
        should be ``typing.Callable[[UIWidget, int], None]``.

        Keyword Args:
            name: handler property in which to store widget (optional)
            items: list combo box items (strings, optional)
            items_ref: handler reference of combo box items (bindable, optional)
            current_index: current index handler reference (bindable, optional)
            on_current_index_changed: callback when current index changes (optional)

        Returns:
            UI description of the combo box
        """
        d = {"type": "combo_box"}
        if name is not None:
            d["name"] = name
        if items is not None:
            d["items"] = items
        if items_ref is not None:
            d["items_ref"] = items_ref
        if current_index is not None:
            d["current_index"] = current_index
        if on_current_index_changed is not None:
            d["on_current_index_changed"] = on_current_index_changed
        self.__process_common_properties(d, **kwargs)
        return d

    def create_radio_button(self, *,
                            name: UIIdentifier=None,
                            text: UILabel=None,
                            value: typing.Any=None,
                            group_value: UIIdentifier=None,
                            **kwargs) -> UIDescription:
        """Create a radio button UI description with text, name, value, and group value.

        A set of radio buttons should be created such that each has a different ``value`` but shares a common
        ``group_value``. The type of ``value`` must match the type of ``group_value``.

        Keyword Args:
            name: handler property in which to store widget (optional)
            text: text of the label (bindable)
            value: unique value within its group (required)
            group_value: common value handler reference (bindable, required)

        Returns:
            UI description of the radio button
        """
        d = {"type": "radio_button"}
        if name is not None:
            d["name"] = name
        if text is not None:
            d["text"] = text
        if value is not None:
            d["value"] = value
        if group_value is not None:
            d["group_value"] = group_value
        self.__process_common_properties(d, **kwargs)
        return d

    def create_slider(self, *,
                      name: UIIdentifier=None,
                      value: UIIdentifier=None,
                      minimum: int=None,
                      maximum: int=None,
                      on_value_changed: UICallableIdentifier=None,
                      on_slider_pressed: UICallableIdentifier=None,
                      on_slider_released: UICallableIdentifier=None,
                      on_slider_moved: UICallableIdentifier=None,
                      **kwargs) -> UIDescription:
        """Create a slider UI description with name, value, limits, and events.

        The ``on_value_changed`` callback is invoked whenever the slider value changes, including if set
        programmatically. The widget and the new value are passed to the callback. The type signature in the handler
        should be ``typing.Callable[[UIWidget, int], None]``.

        The ``on_slider_pressed`` callback is invoked when the user begins dragging the slider. The widget is passed to
        the callback. The type signature in the handler should be ``typing.Callable[[UIWidget], None]``.

        The ``on_slider_released`` callback is invoked when the user stops dragging the slider. The widget is passed to
        the callback. The type signature in the handler should be ``typing.Callable[[UIWidget], None]``.

        The ``on_slider_moved`` callback is invoked whenever the slider value changes while the user is dragging. The
        widget and the new value are passed to the callback. The type signature in the handler should be
        ``typing.Callable[[UIWidget, int], None]``.

        Keyword Args:
            name: handler property in which to store widget (optional)
            value: handler reference to the current value (required, bindable)
            minimum: minimum value (default 0)
            maximum: maximum value (default 100)
            on_value_changed: callback when value changes, any source (optional)
            on_slider_pressed: callback when slider is pressed (optional)
            on_slider_released: callback when slider is released (optional)
            on_slider_moved: callback when slider moves, user initiated only (optional)

        Returns:
            UI description of the slider
        """
        d = {"type": "slider"}
        if name is not None:
            d["name"] = name
        if value is not None:
            d["value"] = value
        if minimum is not None:
            d["minimum"] = minimum
        if maximum is not None:
            d["maximum"] = maximum
        if on_value_changed is not None:
            d["on_value_changed"] = on_value_changed
        if on_slider_pressed is not None:
            d["on_slider_pressed"] = on_slider_pressed
        if on_slider_released is not None:
            d["on_slider_released"] = on_slider_released
        if on_slider_moved is not None:
            d["on_slider_moved"] = on_slider_moved
        self.__process_common_properties(d, **kwargs)
        return d

    def create_progress_bar(self, *,
                            name: UIIdentifier = None,
                            value: UIIdentifier = None,
                            minimum: int = None,
                            maximum: int = None,
                            **kwargs) -> UIDescription:
        """Create a progress bar UI description with name, value, and limits.

        Keyword Args:
            name: handler property in which to store widget (optional)
            value: handler reference to the current value (required, bindable)
            minimum: minimum value (default 0)
            maximum: maximum value (default 100)

        Returns:
            UI description of the progress bar
        """
        d = {"type": "progress_bar"}
        if name is not None:
            d["name"] = name
        if value is not None:
            d["value"] = value
        if minimum is not None:
            d["minimum"] = minimum
        if maximum is not None:
            d["maximum"] = maximum
        self.__process_common_properties(d, **kwargs)
        return d

    def create_modeless_dialog(self, content: UIDescription, *, title: str=None, resources: UIResources=None, margin: UIPoints=None) -> UIDescription:
        """Create a modeless dialog UI description with content, title, resources, and margin.

        Args:
            content: UI description of the content

        Keyword Args:
            title: title of the window
            resources: additional resources
            margin: margin in points

        Returns:
            a UI description of the dialog
        """
        d = {"type": "modeless_dialog", "content": content}
        if title is not None:
            d["title"] = title
        if margin is not None:
            d["margin"] = margin
        if resources is not None:
            d["resources"] = resources
        return d

    def create_window(self, content: UIDescription, *, title: str=None, resources: UIResources=None, margin: UIPoints=None) -> UIDescription:
        """Create a window UI description with content, title, resources, and margin.

        Args:
            content: UI description description of the content

        Keyword Args:
            title: title of the window
            resources: additional resources
            margin: margin in points

        Returns:
            a UI description of the window
        """
        d = {"type": "window", "content": content}
        if title is not None:
            d["title"] = title
        if margin is not None:
            d["margin"] = margin
        if resources is not None:
            d["resources"] = resources
        return d

    def define_component(self, content, *, component_id=None, events=None):
        d = {"type": "component", "content": content}
        if component_id is not None:
            d["component_id"] = component_id
        if events is not None:
            d["events"] = events
        return d

    def create_component_instance(self, identifier, properties=None, **kwargs):
        properties = properties if properties is not None else dict()
        d = {"type": "component", "identifier": identifier, "properties": properties}
        for k, v in kwargs.items():
            d[k] = v
        return d


def connect_name(widget, d, handler):
    name = d.get("name", None)
    if name:
        setattr(handler, name, widget)


def connect_string_value(widget, d, handler, property, finishes):
    """Connects a value in the property, but also allows binding.

    A value means the value for the property is directly contained in the string.
    """
    v = d.get(property)
    m = re.match("^@binding\((.+)\)$", v if v else "")
    # print(f"{v}, {m}, {m.group(1) if m else 'NA'}")
    if m:
        b = m.group(1)
        parts = [p.strip() for p in b.split(',')]
        def finish_binding():
            handler_property_path = parts[0].split('.')
            source = handler
            for p in handler_property_path[:-1]:
                source = getattr(source, p.strip())
            converter = None
            for part in parts:
                if part.startswith("converter="):
                    converter = getattr(handler, part[len("converter="):])
            if hasattr(source, "property_changed_event"):
                binding = Binding.PropertyBinding(source, handler_property_path[-1].strip(), converter=converter)
                getattr(widget, "bind_" + property)(binding)
            else:
                setattr(widget, property, getattr(source, handler_property_path[-1].strip()))
        finishes.append(finish_binding)
    else:
        setattr(widget, property, v)


class Closer:
    """A helper class to facilitate closing handlers and associated closeable items.

    A closer is attached to each handler and used to close the handler, extra closeable items that the engine may
    created, and child component handlers.
    """
    def __init__(self):
        self.__handlers = set()

    def push_closeable(self, handler):
        assert handler not in self.__handlers
        self.__handlers.add(handler)

    def pop_closeable(self, handler):
        assert handler in self.__handlers
        if callable(getattr(handler, "close", None)):
            handler.close()
        if hasattr(handler, "_closer"):
            handler._closer.close()
        self.__handlers.remove(handler)

    def close(self):
        for handler in self.__handlers:
            if callable(getattr(handler, "close", None)):
                handler.close()
            if hasattr(handler, "_closer"):
                handler._closer.close()
        self.__handlers = None


def connect_reference_value(widget, d, handler, property, finishes, binding_name=None, value_type=None):
    """Connects a reference to the property, but also allows binding.

    A reference means the property specifies a property in the handler.
    """
    binding_name = binding_name if binding_name else property
    v = d.get(property)
    m = re.match("^@binding\((.+)\)$", v if v else "")
    # print(f"{v}, {m}, {m.group(1) if m else 'NA'}")
    if m:
        b = m.group(1)
        parts = [p.strip() for p in b.split(',')]

        # finish binding is called after the window has been constructed using the 'finishes' list.
        def finish_binding():
            handler_property_path = parts[0].split('.')
            source = handler
            for p in handler_property_path[:-1]:
                source = getattr(source, p.strip())
            converter = None
            # check if any of the parts has a converter.
            for part in parts:
                if part.startswith("converter="):
                    converter = getattr(handler, part[len("converter="):])
            # give the handler a chance to make object conversions. this is useful if the objects
            # in the handler are stored in a proxy format or something similar.
            if getattr(handler, "get_object_converter", None):
                converter = handler.get_object_converter(converter)
            # configure the binding if the source and widget meet the criteria.
            if hasattr(source, "property_changed_event") and hasattr(widget, "bind_" + binding_name):
                binding = Binding.PropertyBinding(source, handler_property_path[-1].strip(), converter=converter)
                getattr(widget, "bind_" + binding_name)(binding)
            # otherwise just set the value.
            else:
                setattr(widget, binding_name, getattr(source, handler_property_path[-1].strip()))

        finishes.append(finish_binding)
    elif v is not None:
        if value_type == str and hasattr(handler, v):
            # backwards compatible binding
            setattr(widget, binding_name, getattr(handler, v))
        elif value_type and isinstance(v, value_type):
            setattr(widget, binding_name, v)
        else:
            setattr(widget, binding_name, getattr(handler, v))


def connect_event(widget, source, d, handler, event_str, arg_names):
    event_method_name = d.get(event_str, None)
    if event_method_name:
        event_fn = getattr(handler, event_method_name)
        if event_fn:
            def trampoline(*args, **kwargs):
                combined_args = dict()
                for arg_name, arg in zip(arg_names, args):
                    combined_args[arg_name] = arg
                combined_args.update(kwargs)
                return event_fn(widget, **combined_args)
            setattr(source, event_str, trampoline)
        else:
            print("WARNING: '" + event_str + "' method " + event_method_name + " not found in handler.")


def connect_attributes(widget, d, handler, finishes):
    connect_reference_value(widget, d, handler, "enabled", finishes, value_type=bool)
    connect_reference_value(widget, d, handler, "visible", finishes, value_type=bool)
    connect_string_value(widget, d, handler, "tool_tip", finishes)


def run_window(app, d, handler):
    ui = app.ui
    d_type = d.get("type")
    if d_type == "window":
        title = d.get("title", _("Untitled"))
        margin = d.get("margin")
        persistent_id = d.get("persistent_id")
        content = d.get("content")
        resources = d.get("resources", dict())
        for k, v in resources.items():
            resources[k] = v
        if not hasattr(handler, "resources"):
            handler.resources = resources
        else:
            handler.resources.update(resources)
        closer = Closer()
        finishes = list()
        window = Window.Window(ui, app=app, persistent_id=persistent_id)
        window.title = title
        window.on_close = closer.close
        # make and attach closer for the handler; put handler into container closer
        handler._closer = Closer()
        closer.push_closeable(handler)
        outer_row = ui.create_row_widget()
        outer_column = ui.create_column_widget()
        inner_content = construct(ui, window, content, handler, finishes)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        outer_column.add(inner_content)
        outer_row.add(outer_column)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        window.attach_widget(outer_row)
        window.show()
        for finish in finishes:
            finish()
        handler._event_loop = window.event_loop
        if callable(getattr(handler, "init_handler", None)):
            handler.init_handler()
        return window


def construct_margin(ui, content, margin):
    if margin:
        column = ui.create_column_widget()
        column.add_spacing(margin)
        column.add(content)
        column.add_spacing(margin)
        row = ui.create_row_widget()
        row.add_spacing(margin)
        row.add(column)
        row.add_spacing(margin)
        content = row
    return content


def connect_items(ui, window, container_widget, handler, items, item_component_id, finishes):
    assert window is not None
    items_parts = items.split('.')
    container = handler
    for items_part in items_parts[:-1]:
        container = getattr(container, items_part.strip())
    items_key = items_parts[-1]

    # the _closer should have been set on the handler, even if no close method is present. insert_item makes this
    # assumption so that sub-components have a path by which to get closed.
    assert handler._closer

    def insert_item(index, item):
        item_widget = None
        component = handler.resources.get(item_component_id)
        if component:
            assert component.get("type") == "component"
            # the component will have a content portion, which is a widget description. component events are
            # ignored in this case.
            content = component.get("content")
            component_id = component.get("component_id")
            assert component_id == item_component_id
            assert callable(getattr(handler, "create_handler", None))
            # create the handler first, but don't initialize it.
            component_handler = handler.create_handler(component_id=component_id, item=item, container=container)
            # make and attach closer for the component handler and link it to the container handler.
            if component_handler:
                component_handler._closer = Closer()
                handler._closer.push_closeable(component_handler)
            item_finishes = list()
            # now construct the widget
            item_widget = construct(ui, window, content, component_handler, item_finishes)
            # since the handler is custom to the widget, make a way to retrieve it from the widget
            item_widget.handler = component_handler
            for finish in item_finishes:
                finish()
            component_handler._event_loop = window.event_loop
            if callable(getattr(component_handler, "init_handler", None)):
                component_handler.init_handler()
        container_widget.insert(item_widget, index)
        return item_widget

    def row_item_inserted(key, value, before_index):
        if key == items_key:
            insert_item(before_index, value)

    def row_item_removed(key, value, before_index):
         if key == items_key:
            item_widget = container_widget.children[before_index]
            handler._closer.pop_closeable(item_widget.handler)
            container_widget.remove(item_widget)

    for item in getattr(container, items_key):
        insert_item(len(container_widget.children), item)

    handler._closer.push_closeable(container.item_inserted_event.listen(row_item_inserted))
    handler._closer.push_closeable(container.item_removed_event.listen(row_item_removed))


def construct_sizing_properties(d: typing.Mapping) -> typing.Dict:
    properties = dict()
    for k in ("width", "min_width", "max_width", "height", "min_height", "max_height"):
        v = d.get(k, None)
        if v is not None:
            properties[k] = int(v)
    return properties


class DeclarativeConstructor:
    def construct(self, d_type: str, ui: UserInterface.UserInterface, window, d: typing.Mapping, handler, finishes: typing.Sequence[typing.Callable[[], None]] = None): ...


def construct(ui: UserInterface.UserInterface, window: Window.Window, d: typing.Mapping, handler, finishes: typing.Sequence[typing.Callable[[], None]] = None):
    d_type = d.get("type")
    if d_type == "modeless_dialog":
        title = d.get("title", _("Untitled"))
        margin = d.get("margin")
        persistent_id = d.get("persistent_id")
        content = d.get("content")
        resources = d.get("resources", dict())
        for k, v in resources.items():
            resources[k] = v
        if not hasattr(handler, "resources"):
            handler.resources = resources
        else:
            handler.resources.update(resources)
        closer = Closer()
        finishes = list()
        dialog = Dialog.ActionDialog(ui, title, app=window.app, parent_window=window, persistent_id=persistent_id)
        dialog.on_close = closer.close
        dialog._create_menus()
        # make and attach closer for the handler; put handler into container closer
        handler._closer = Closer()
        closer.push_closeable(handler)
        outer_row = ui.create_row_widget()
        outer_column = ui.create_column_widget()
        inner_content = construct(ui, window, content, handler, finishes)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        outer_column.add(inner_content)
        outer_row.add(outer_column)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        dialog.content.add(outer_row)
        for finish in finishes:
            finish()
        handler._event_loop = window.event_loop
        if callable(getattr(handler, "init_handler", None)):
            handler.init_handler()
        return dialog
    elif d_type == "column":
        properties = construct_sizing_properties(d)
        column_widget = ui.create_column_widget(properties=properties)
        spacing = d.get("spacing")
        margin = d.get("margin")
        items = d.get("items")
        item_component_id = d.get("item_component_id")
        children = d.get("children", list())
        assert not items or not children
        first = True
        for child in children:
            if not first and spacing is not None:
                column_widget.add_spacing(spacing)
            if child.get("type") == "spacing":
                column_widget.add_spacing(child.get("size", 0))
            elif child.get("type") == "stretch":
                column_widget.add_stretch()
            else:
                column_widget.add(construct(ui, window, child, handler, finishes))
            first = False
        if items and item_component_id:
            # TODO: spacing does not work on rows/columns
            connect_items(ui, window, column_widget, handler, items, item_component_id, finishes)
        if handler:
            connect_attributes(column_widget, d, handler, finishes)
        return construct_margin(ui, column_widget, margin)
    elif d_type == "row":
        properties = construct_sizing_properties(d)
        row_widget = ui.create_row_widget(properties=properties)
        spacing = d.get("spacing")
        margin = d.get("margin")
        items = d.get("items")
        item_component_id = d.get("item_component_id")
        children = d.get("children", list())
        assert not items or not children
        first = True
        for child in children:
            if not first and spacing is not None:
                row_widget.add_spacing(spacing)
            if child.get("type") == "spacing":
                row_widget.add_spacing(child.get("size", 0))
            elif child.get("type") == "stretch":
                row_widget.add_stretch()
            else:
                row_widget.add(construct(ui, window, child, handler, finishes))
            first = False
        if items and item_component_id:
            # TODO: spacing does not work on rows/columns
            connect_items(ui, window, row_widget, handler, items, item_component_id, finishes)
        if handler:
            connect_attributes(row_widget, d, handler, finishes)
        return construct_margin(ui, row_widget, margin)
    elif d_type == "text_label":
        properties = construct_sizing_properties(d)
        widget = ui.create_label_widget(None, properties)
        if handler:
            connect_string_value(widget, d, handler, "text", finishes)
            connect_name(widget, d, handler)
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "line_edit":
        editable = d.get("editable", None)
        clear_button_enabled = d.get("clear_button_enabled", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_line_edit_widget(properties)
        if editable is not None:
            widget.editable = editable
        if clear_button_enabled is not None:
            widget.clear_button_enabled = clear_button_enabled
        if handler:
            connect_name(widget, d, handler)
            connect_string_value(widget, d, handler, "placeholder_text", finishes)
            connect_reference_value(widget, d, handler, "text", finishes)
            connect_event(widget, widget, d, handler, "on_editing_finished", ["text"])
            connect_event(widget, widget, d, handler, "on_escape_pressed", [])
            connect_event(widget, widget, d, handler, "on_return_pressed", [])
            connect_event(widget, widget, d, handler, "on_key_pressed", ["key"])
            connect_event(widget, widget, d, handler, "on_text_edited", ["text"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "text_edit":
        editable = d.get("editable", None)
        clear_button_enabled = d.get("clear_button_enabled", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_text_edit_widget(properties)
        if editable is not None:
            widget.editable = editable
        if clear_button_enabled is not None:
            widget.clear_button_enabled = clear_button_enabled
        if handler:
            connect_name(widget, d, handler)
            connect_string_value(widget, d, handler, "placeholder_text", finishes)
            connect_reference_value(widget, d, handler, "text", finishes)
            connect_event(widget, widget, d, handler, "on_escape_pressed", [])
            connect_event(widget, widget, d, handler, "on_return_pressed", [])
            connect_event(widget, widget, d, handler, "on_text_edited", ["text"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "push_button":
        text = d.get("text", None)
        icon_identifier = d.get("icon", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_push_button_widget(text, properties)
        if handler:
            if icon_identifier:
                connect_reference_value(widget, d, handler, "icon", finishes)
            connect_name(widget, d, handler)
            connect_event(widget, widget, d, handler, "on_clicked", [])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "check_box":
        # TODO: 'checked' and 'check_state' are bindings, not values
        text = d.get("text", None)
        tristate = d.get("tristate", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_check_box_widget(text, properties=properties)
        if tristate is not None:
            widget.tristate = tristate
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "checked", finishes, value_type=bool)
            connect_reference_value(widget, d, handler, "check_state", finishes, value_type=str)
            connect_event(widget, widget, d, handler, "on_checked_changed", ["checked"])
            connect_event(widget, widget, d, handler, "on_check_state_changed", ["check_state"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "combo_box":
        items = d.get("items", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_combo_box_widget(items=items, properties=properties)
        if handler:
            connect_name(widget, d, handler)
            # note: items_ref connects before current_index so that current_index can be valid
            connect_reference_value(widget, d, handler, "items_ref", finishes, binding_name="items", value_type=list)
            connect_reference_value(widget, d, handler, "current_index", finishes, value_type=int)
            connect_event(widget, widget, d, handler, "on_current_index_changed", ["current_index"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "radio_button":
        text = d.get("text", None)
        value = d.get("value", None)
        properties = construct_sizing_properties(d)
        widget = ui.create_radio_button_widget(text, properties)
        widget.value = value
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "group_value", finishes, value_type=int)
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "slider":
        minimum = d.get("minimum", 0)
        maximum = d.get("maximum", 100)
        properties = construct_sizing_properties(d)
        widget = ui.create_slider_widget(properties)
        widget.minimum = minimum
        widget.maximum = maximum
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "value", finishes, value_type=int)
            connect_event(widget, widget, d, handler, "on_value_changed", ["value"])
            connect_event(widget, widget, d, handler, "on_slider_pressed", [])
            connect_event(widget, widget, d, handler, "on_slider_released", [])
            connect_event(widget, widget, d, handler, "on_slider_moved", ["value"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "progress_bar":
        minimum = d.get("minimum", 0)
        maximum = d.get("maximum", 100)
        properties = construct_sizing_properties(d)
        properties.setdefault("height", 18)
        properties.setdefault("width", 64)
        widget = ui.create_progress_bar_widget(properties=properties)
        widget.minimum = minimum
        widget.maximum = maximum
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "value", finishes, value_type=int)
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "tabs":
        properties = construct_sizing_properties(d)
        widget = ui.create_tab_widget(properties)
        for tab in d.get("tabs", list()):
            widget.add(construct(ui, window, tab["content"], handler, finishes), tab["label"])
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "current_index", finishes, value_type=int)
            connect_event(widget, widget, d, handler, "on_current_index_changed", ["current_index"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "stack":
        properties = construct_sizing_properties(d)
        widget = ui.create_stack_widget(properties)
        for child in d.get("children", list()):
            widget.add(construct(ui, window, child, handler, finishes))
        items = d.get("items")
        item_component_id = d.get("item_component_id")
        if items and item_component_id:
            connect_items(ui, window, widget, handler, items, item_component_id, finishes)
        if handler:
            connect_name(widget, d, handler)
            connect_reference_value(widget, d, handler, "current_index", finishes, value_type=int)
            connect_event(widget, widget, d, handler, "on_current_index_changed", ["current_index"])
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "scroll_area":
        properties = construct_sizing_properties(d)
        widget = ui.create_scroll_area_widget(properties)
        widget.set_scrollbar_policies("needed", "needed")
        content = d.get("content")
        widget.content = construct(ui, window, content, handler, finishes)
        if handler:
            connect_name(widget, d, handler)
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "group":
        properties = construct_sizing_properties(d)
        widget = ui.create_group_widget(properties)
        margin = d.get("margin")
        content = d.get("content")
        outer_row = ui.create_row_widget()
        outer_column = ui.create_column_widget()
        inner_content = construct(ui, window, content, handler, finishes)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        outer_column.add(inner_content)
        outer_row.add(outer_column)
        if margin is not None:
            outer_row.add_spacing(margin)
            outer_column.add_spacing(margin)
        widget.add(outer_row)
        if handler:
            connect_name(widget, d, handler)
            connect_string_value(widget, d, handler, "title", finishes)
            connect_attributes(widget, d, handler, finishes)
        return widget
    elif d_type == "component":
        # a component needs to be registered before it is instantiated.
        # look up the identifier in the handler resoureces.
        identifier = d.get("identifier", None)
        component = handler.resources.get(identifier)
        if component:
            assert component.get("type") == "component"
            # the component will have a content portion, which is a widget description, and a list of events.
            content = component.get("content")
            component_id = component.get("component_id")
            events = component.get("events", list())
            # create the handler first, but don't initialize it.
            component_handler = handler.create_handler(component_id=component_id) if component_id and hasattr(handler, "create_handler") else None
            if component_handler:
                # make and attach closer for the component handler and link it to the container handler.
                component_handler._closer = Closer()
                handler._closer.push_closeable(component_handler)
                # set properties in the component from the properties dict
                for k, v in d.get("properties", dict()).items():
                    # print(f"setting property {k} to {v}")
                    setattr(component_handler, k, v)
            # now construct the widget
            widget = construct(ui, window, content, component_handler, finishes)
            # connect the name to the handler if desired
            connect_name(widget, d, handler)
            # since the handler is custom to the widget, make a way to retrieve it from the widget
            widget.handler = component_handler
            component_handler._event_loop = window.event_loop
            if callable(getattr(component_handler, "init_handler", None)):
                component_handler.init_handler()
            # connect events
            for event in events:
                # print(f"connecting {event['event']} ({event['parameters']})")
                connect_event(widget, component_handler, d, handler, event["event"], event["parameters"])
            if handler:
                connect_attributes(widget, d, handler, finishes)
            return widget
    else:
        # if the component is not handled yet, check with registered component handlers.
        constructors = typing.cast(typing.List[DeclarativeConstructor], Registry.get_components_by_type("declarative_constructor"))
        for constructor in constructors:
            widget = constructor.construct(d_type, ui, window, d, handler, finishes)
            if widget:
                return widget
    return None


class DeclarativeWidget(Widgets.CompositeWidgetBase):
    """A widget containing a declarative ui handler."""

    def __init__(self, ui, event_loop, ui_handler):
        super().__init__(ui.create_stack_widget())

        # create a top level closer. for each object added to a closer, the closer will
        # call close (if it exists) and then close the object's _closer (if it exists).
        self.__closer = Closer()
        if ui_handler:
            # create a _closer and attach it to the ui_handler. this may be used for sub-components.
            # then add the ui_handler to itself to be closed by the top level closer.
            ui_handler._closer = Closer()
            self.__closer.push_closeable(ui_handler)

        class Window:
            # dummy Window to supply event loop
            def __init__(self):
                self.event_loop = event_loop

        finishes = list()
        widget = construct(ui, Window(), ui_handler.ui_view, ui_handler, finishes)
        self.content_widget.add(widget)
        for finish in finishes:
            finish()
        ui_handler._event_loop = event_loop
        if callable(getattr(ui_handler, "init_handler", None)):
            ui_handler.init_handler()

    def close(self):
        self.__closer.close()
        super().close()


def run_ui(args, bootstrap_args, d, handler):

    def start():
        run_window(app, d, handler)
        return True

    app = Application.Application(Application.make_ui(bootstrap_args), on_start=start)
    app.initialize()
    return app



import numpy

from nion.ui import Declarative
from nion.utils import Event


def make_icon(w: int, h: int) -> numpy.ndarray:
    # make bitmap_data (random static) for icon push button
    bitmap = numpy.zeros((h, w, 4), numpy.uint8)
    bitmap[..., 0] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # blue
    bitmap[..., 1] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # green
    bitmap[..., 2] = (numpy.random.randn(h, w) * 255).astype(numpy.uint8)  # red
    bitmap[..., 3] = 255
    return bitmap.view(numpy.uint32).reshape(bitmap.shape[:-1])


class Handler:

    def __init__(self):
        self.icon1 =  make_icon(12, 12)
        self.icon2 =  make_icon(48, 36)
        self.property_changed_event = Event.Event()

    def click(self, widget):
        self.icon1 =  make_icon(12, 12)
        self.property_changed_event.fire("icon1")


def construct_ui(ui: Declarative.DeclarativeUI) -> Declarative.UIDescription:
    text_only_button = ui.create_push_button(text="Change Button Icon", on_clicked="click")
    icon_only_button = ui.create_push_button(icon="@binding(icon2)")
    text_and_icon_button = ui.create_push_button(text="Text", icon="@binding(icon1)")

    hellos = ui.create_column(
        ui.create_row(text_only_button, ui.create_stretch()),
        ui.create_row(icon_only_button, ui.create_stretch()),
        ui.create_row(text_and_icon_button, ui.create_stretch()),
        ui.create_stretch(),
        spacing=12)

    return hellos



import typing

from nion.ui import Declarative
from nion.utils import Event
from nion.utils import Model

class Section:

    def __init__(self, title: str):
        self.title = title
        self.count = 0


class Handler:

    def __init__(self):
        self.sections = list()
        self.item_inserted_event = Event.Event()
        self.item_removed_event = Event.Event()
        self.sections.append(Section("Apples"))
        self.sections.append(Section("Oranges"))
        self.title_model = Model.PropertyModel()

    def add(self, widget) -> None:
        title = self.title_model.value
        if title:
            section = Section(title)
            index = len(self.sections)
            self.sections.append(section)
            self.item_inserted_event.fire("sections", section, index)

    def remove(self, section: Section) -> None:
        index = self.sections.index(section)
        del self.sections[index]
        self.item_removed_event.fire("sections", section, index)

    def create_handler(self, component_id: str, container=None, item=None, **kwargs):

        class SectionHandler:

            def __init__(self, container: typing.List[Section], section: Section):
                self.container = container
                self.section = section
                self.title_label_widget = None
                self.count_label_widget = None

            def init_handler(self):
                # when this is called, all fields will be populated
                self.title_label_widget.text = self.section.title
                self.count_label_widget.text = str(self.section.count)

            def increase_count(self, widget):
                self.section.count += 1
                self.count_label_widget.text = str(self.section.count)

            def decrease_count(self, widget):
                self.section.count -= 1
                self.count_label_widget.text = str(self.section.count)

            def remove(self, widget):
                self.container.remove(self.section)

        if component_id == "section":
            return SectionHandler(container, item)

    @property
    def resources(self):
        ui = Declarative.DeclarativeUI()
        title_label = ui.create_label(name="title_label_widget")
        count_label = ui.create_label(name="count_label_widget")
        increase_button = ui.create_push_button(text="++", on_clicked="increase_count")
        decrease_button = ui.create_push_button(text="--", on_clicked="decrease_count")
        remove_button = ui.create_push_button(text="X", on_clicked="remove")
        row = ui.create_row(title_label, count_label, increase_button, decrease_button, remove_button, ui.create_stretch(), spacing=8)
        component = ui.define_component(content=row, component_id="section")
        return {"section": component}


def construct_ui(ui: Declarative.DeclarativeUI):
    title_field = ui.create_line_edit(text="@binding(title_model.value)")
    add_button = ui.create_push_button(text="Add", on_clicked="add")
    sections_column = ui.create_column(items="sections", item_component_id="section", spacing=8)
    return ui.create_column(title_field, add_button, sections_column, spacing=8, margin=12)



"""Display a tree of drawable cells.
"""

# standard libraries
import copy
import functools
import json

# third party libraries
# none

# local libraries
from . import CanvasItem
from nion.utils import Geometry


class TreeCanvasItem(CanvasItem.CanvasItemComposition):
    """
    Takes a delegate that supports the following properties, methods, and optional methods:

    Properties:
        None

    Methods:
        toggle_is_expanded(value_path) -> None
        build_items(get_font_metrics_fn, item_width) -> CanvasItem

    Optional methods:
        None

    Call reconstruct when data or selection changes.
    """

    def __init__(self, get_font_metrics_fn, delegate):
        super().__init__()
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__delegate = delegate
        # configure super
        self.wants_mouse_events = True
        self.focusable = True
        # internal variables
        self.__mouse_pressed = False
        self.__mouse_index = None
        self.__mouse_position = None
        self.__mouse_dragging = False
        self.__mouse_item = None
        self.__selected_value_paths = set()
        self.layout = CanvasItem.CanvasItemColumnLayout()
        self.on_content_height_changed = None

    def close(self):
        self.on_content_height_changed = None
        super().close()

    def __is_selected(self, value_path):
        return json.dumps(value_path) in self.__selected_value_paths

    def reconstruct(self):
        for canvas_item in copy.copy(self.canvas_items):
            self._remove_canvas_item(canvas_item)
        indent_size = 16
        canvas_bounds = self.canvas_bounds
        item_width = int(canvas_bounds.width) if canvas_bounds else None
        canvas_height = 0
        ITEM_HEIGHT = 18
        for canvas_item, item_type, is_expanded, value_path in self.__delegate.build_items(
                self.__get_font_metrics_fn, item_width):
            indent = (len(value_path) - 1) * indent_size
            item_row = CanvasItem.CanvasItemComposition()
            item_row.sizing.set_fixed_height(ITEM_HEIGHT)
            item_row.layout = CanvasItem.CanvasItemRowLayout()
            item_row.add_spacing(indent)
            if item_type == "parent":
                twist_down_canvas_item = CanvasItem.TwistDownCanvasItem()
                twist_down_canvas_item.sizing.set_fixed_size(Geometry.IntSize(height=ITEM_HEIGHT, width=16))
                twist_down_canvas_item.checked = is_expanded

                def twist_down_clicked(toggle_value_path):
                    self.__toggle_is_expanded(toggle_value_path)

                twist_down_canvas_item.on_button_clicked = functools.partial(twist_down_clicked, value_path)
                item_row.add_canvas_item(twist_down_canvas_item)
            else:
                item_row.add_spacing(indent_size)
            item_row.add_canvas_item(canvas_item)
            item_row.add_stretch()
            item_row.value_path = value_path
            self.add_canvas_item(item_row)
            canvas_height += ITEM_HEIGHT
        self.update()
        if callable(self.on_content_height_changed):
            self.on_content_height_changed(canvas_height)

    def __set_selection(self, value_path):
        self.__selected_value_paths.clear()
        self.__selected_value_paths.add(json.dumps(value_path))

    def __extend_selection(self, value_path):
        pass

    def __toggle_selection(self, value_path):
        value_path_key = json.dumps(value_path)
        if value_path_key in self.__selected_value_paths:
            self.__selected_value_paths.remove(value_path_key)
        else:
            self.__selected_value_paths.add(value_path_key)

    def __toggle_is_expanded(self, value_path):
        self.__delegate.toggle_is_expanded(value_path)
        self.reconstruct()

    def __context_menu_event(self, value_path, x, y, gx, gy):
        pass

    def __drag_started(self, value_path, x, y, modifiers):
        pass

    def __delete(self):
        pass

    def __adjust_selection(self, action, extend):
        pass

    def __value_path_at_point(self, p):
        for canvas_item in self.canvas_items_at_point(p.x, p.y):
            if hasattr(canvas_item, "value_path"):
                return canvas_item.value_path
        return None

    def context_menu_event(self, x, y, gx, gy):
        p = Geometry.IntPoint(y=y, x=x)
        value_path = self.__value_path_at_point(p)
        if value_path:
            if not self.__is_selected(value_path):
                self.__set_selection(value_path)
            return self.__context_menu_event(value_path, x, y, gx, gy)
        return self.__context_menu_event(None, x, y, gx, gy)

    def mouse_pressed(self, x, y, modifiers):
        p = Geometry.IntPoint(y=y, x=x)
        value_path = self.__value_path_at_point(p)
        if value_path:
            if modifiers.shift:
                self.__extend_selection(value_path)
            elif modifiers.control:
                self.__toggle_selection(value_path)
            else:
                self.__set_selection(value_path)
                self.__mouse_pressed = True
                self.__mouse_position = Geometry.IntPoint(y=y, x=x)
                self.__mouse_item = value_path
            return True
        return super().mouse_pressed(x, y, modifiers)

    def mouse_released(self, x, y, modifiers):
        if self.__mouse_pressed:
            # double check whether mouse_released has been called explicitly as part of a drag.
            # see https://bugreports.qt.io/browse/QTBUG-40733
            pass  # leave this here for future reference
        self.__mouse_pressed = False
        self.__mouse_item = None
        self.__mouse_position = None
        self.__mouse_dragging = False
        return True

    def mouse_position_changed(self, x, y, modifiers):
        if self.__mouse_pressed:
            if not self.__mouse_dragging and Geometry.distance(self.__mouse_position, Geometry.IntPoint(y=y, x=x)) > 8:
                self.__mouse_dragging = True
                self.__drag_started(self.__mouse_item, x, y, modifiers)
                # once a drag starts, mouse release will not be called; call it here instead
                self.mouse_released(x, y, modifiers)
                return True
        return super().mouse_position_changed(x, y, modifiers)

    def key_pressed(self, key):
        if key.is_delete:
            self.__delete()
            return True
        if key.is_up_arrow:
            self.__adjust_selection("up", key.modifiers.shift)
            return True
        if key.is_down_arrow:
            self.__adjust_selection("down", key.modifiers.shift)
            return True
        return super().key_pressed(key)

    def handle_select_all(self):
        self.__adjust_selection("all", False)
        return True



# -*- coding: utf-8 -*-

import setuptools
import os

# for some reason os gets munged after this point on Windows, so compute it here.
readme_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md")

setuptools.setup(
    name="nionui",
    version="0.3.26",
    author="Nion Software",
    author_email="swift@nion.com",
    description="Nion UI framework.",
    long_description=open(readme_path).read(),
    url="https://github.com/nion-software/nionui",
    packages=["nion.ui", "nion.ui.test", "nionui_app.none", "nionui_app.nionui_examples.hello_world", "nionui_app.nionui_examples.ui_demo"],
    install_requires=['numpy', 'nionutils>=0.3.19', 'imageio'],
    license='Apache 2.0',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
    ],
    include_package_data=True,
    package_data={
        'nion.ui': ['*.html', "resources/*"],
    },
    test_suite="nion.ui.test",
    entry_points={
        'console_scripts': [
            'nionui=nion.ui.command:main',
            ],
        },
)



from nion.utils import Model

class Handler:

    def __init__(self):
        self.cb1 = None
        self.cb2_current_index_model = Model.PropertyModel(1)
        self.cb2_current_index_model.on_value_changed = self.cb2_current_index_changed
        self.numbers = ["One", "Two", "Three"]
        self.numeros = Model.PropertyModel([])
        self.numeros.value = ["Uno", "Dos", "Tres"]

    def init_handler(self):
        # this method is called after all fields are populated
        self.cb1.current_index = 2

    def cb1_current_index_changed(self, widget, current_index):
        print(f"CB1 {current_index}")

    def cb2_current_index_changed(self, current_index):
        print(f"CB2 {current_index}")

    def change_items(self, widget):
        self.numeros.value = ["Eins", "Zwei"]#, "Drei"]


def construct_ui(ui):
    cb1 = ui.create_combo_box(name="cb1", items=["Red", "Green", "Blue"], on_current_index_changed="cb1_current_index_changed")
    cb2 = ui.create_combo_box(items=["Sheriff", "Astronaut", "Scientist"], current_index="@binding(cb2_current_index_model.value)")
    cb3 = ui.create_combo_box(items_ref="numbers", current_index="@binding(cb2_current_index_model.value)")
    cb4 = ui.create_combo_box(items_ref="@binding(numeros.value)", current_index="@binding(cb2_current_index_model.value)")
    button = ui.create_push_button(text="Change Items", on_clicked="change_items")

    cb_group = ui.create_column(cb1, cb2, cb3, cb4, button, spacing=12)

    return cb_group



"""
Provides a user interface object that can render to an HTML canvas.
"""

# standard libraries
import binascii
import collections
import numbers
import pickle
import queue
import threading
import time

# third party libraries
# None

# local libraries
from . import CanvasItem
from . import DrawingContext
from . import UserInterface
from nion.utils import Geometry


class WidgetBehavior:

    def __init__(self, canvas_item, properties):
        self.properties = properties if properties else {}
        self.canvas_item = canvas_item
        self.__root_container = None
        self.__is_focused = False
        self.__visible = True
        self.__enabled = True
        self.__tool_tip = None
        self.on_context_menu_event = None
        self.on_focus_changed = None

    def close(self):
        self.canvas_item = None

    def _set_root_container(self, root_container):
        self.__root_container = root_container

    @property
    def _root_container(self):
        return self.__root_container

    @property
    def focused(self):
        return self.__root_container.get_is_focused(self.canvas_item)

    @focused.setter
    def focused(self, focused):
        self.__root_container.set_is_focused(self.canvas_item, focused)

    @property
    def visible(self):
        return self.__visible

    @visible.setter
    def visible(self, visible):
        if visible != self.__visible:
            self.__root_container.set_visible(self.canvas_item, visible)
            self.__visible = visible

    @property
    def enabled(self):
        return self.__enabled

    @enabled.setter
    def enabled(self, enabled):
        if enabled != self.__enabled:
            self.__root_container.set_enabled(self.canvas_item, enabled)
            self.__enabled = enabled

    @property
    def size(self):
        return self.__root_container.get_widget_size(self.canvas_item)

    @size.setter
    def size(self, size):
        self.__root_container.set_widget_size(self.canvas_item, size)

    @property
    def tool_tip(self):
        return self.__tool_tip

    @tool_tip.setter
    def tool_tip(self, tool_tip):
        if tool_tip != self.__tool_tip:
            self.__tool_tip = tool_tip

    def handle_drag(self, mime_data: UserInterface.MimeData, thumbnail=None, hot_spot_x=None, hot_spot_y=None, drag_finished_fn=None):
        raise NotImplementedError()

    def handle_context_menu_event(self, x, y, gx, gy):
        if callable(self.on_context_menu_event):
            return self.on_context_menu_event(x, y, gx, gy)
        return False

    def handle_focus_in(self):
        if callable(self.on_focus_changed):
            self.on_focus_changed(True)

    def handle_focus_out(self):
        if callable(self.on_focus_changed):
            self.on_focus_changed(False)

    def map_to_global(self, p):
        return self.canvas_item.map_to_root_container(p)


def extract_canvas_item(widget):
    if hasattr(widget, "content_widget"):
        return extract_canvas_item(widget.content_widget)
    elif hasattr(widget, "_behavior"):
        return widget._behavior.canvas_item
    return None


ChildDescription = collections.namedtuple("ChildDescription", ["widget", "fill", "alignment"])


class BoxWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type, properties):
        super().__init__(CanvasItem.CanvasItemComposition(), properties)
        if widget_type == "row":
            self.canvas_item.layout = CanvasItem.CanvasItemRowLayout()
        else:
            self.canvas_item.layout = CanvasItem.CanvasItemColumnLayout()

    def insert(self, child, index, fill, alignment):
        # behavior must handle index of None, meaning insert at end
        child_canvas_item = extract_canvas_item(child)
        assert child_canvas_item is not None
        index = index if index is not None else self.canvas_item.canvas_items_count
        # TODO: handle fille and alignment
        self.canvas_item.insert_canvas_item(index, extract_canvas_item(child))

    def add_stretch(self) -> UserInterface.Widget:
        return UserInterface.Widget(WidgetBehavior(self.canvas_item.add_stretch(), None))

    def add_spacing(self, spacing: int) -> UserInterface.Widget:
        return UserInterface.Widget(WidgetBehavior(self.canvas_item.add_spacing(spacing), None))


class PushButtonWidgetBehavior(WidgetBehavior):

    def __init__(self, get_font_metrics_fn, properties):
        super().__init__(CanvasItem.TextButtonCanvasItem(), properties)
        self.canvas_item.font = "normal 15px sans-serif"
        self.canvas_item.on_button_clicked = self.__clicked
        self.on_clicked = None
        self.__get_font_metrics_fn = get_font_metrics_fn

    def close(self):
        self.on_clicked = None
        super().close()

    @property
    def text(self):
        return self.canvas_item.text

    @text.setter
    def text(self, text):
        self.canvas_item.text = text
        self.canvas_item.size_to_content(self.__get_font_metrics_fn)

    @property
    def icon(self):
        return self.__icon

    @icon.setter
    def icon(self, rgba_image):  # bgra
        self.__icon = rgba_image
        self.__width = rgba_image.shape[1] if rgba_image is not None else 0
        self.__height = rgba_image.shape[0] if rgba_image is not None else 0

    def __clicked(self):
        if callable(self.on_clicked):
            self.on_clicked()


class CheckBoxWidgetBehavior(WidgetBehavior):

    def __init__(self, get_font_metrics_fn, properties):
        super().__init__(CanvasItem.CheckBoxCanvasItem(), properties)
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.canvas_item.font = "normal 15px sans-serif"
        self.on_check_state_changed = None

        def check_state_changed(check_state):
            if callable(self.on_check_state_changed):
                self.on_check_state_changed(check_state)

        self.canvas_item.on_check_state_changed = check_state_changed

    def close(self):
        self.on_check_state_changed = None
        super().close()

    @property
    def text(self):
        return self.canvas_item.text

    @text.setter
    def text(self, value):
        self.canvas_item.text = value
        self.canvas_item.size_to_content(self.__get_font_metrics_fn)

    @property
    def tristate(self):
        return self.canvas_item.tristate

    @tristate.setter
    def tristate(self, value):
        self.canvas_item.tristate = value

    @property
    def check_state(self):
        return self.canvas_item.check_state

    @check_state.setter
    def check_state(self, value):
        self.canvas_item.check_state = value


class LabelWidgetBehavior(WidgetBehavior):

    def __init__(self, get_font_metrics_fn, properties):
        super().__init__(CanvasItem.StaticTextCanvasItem(), properties)
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__text = None

    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, text):
        self.__text = text if text else ""
        self.canvas_item.text = self.text
        self.canvas_item.size_to_content(self.__get_font_metrics_fn)


class CanvasWidgetBehavior(WidgetBehavior):

    def __init__(self, properties):
        super().__init__(CanvasItem.CanvasItemComposition(), properties)
        self.on_periodic = None
        self.on_mouse_entered = None
        self.on_mouse_exited = None
        self.on_mouse_clicked = None
        self.on_mouse_double_clicked = None
        self.on_mouse_pressed = None
        self.on_mouse_released = None
        self.on_mouse_position_changed = None
        self.on_grabbed_mouse_position_changed = None
        self.on_wheel_changed = None
        self.on_key_pressed = None
        self.on_key_released = None
        self.on_size_changed = None
        self.on_drag_enter = None
        self.on_drag_leave = None
        self.on_drag_move = None
        self.on_drop = None
        self.on_tool_tip = None
        self.on_pan_gesture = None
        self.width = properties.get("width", 0) if properties else 0
        self.height = properties.get("height", 0) if properties else 0
        self.__focusable = False
        self.__draw_mutex = threading.Lock()  # don't delete while drawing
        if self.width > 0:
            self.canvas_item.sizing.set_fixed_width(self.width)
        if self.height > 0:
            self.canvas_item.sizing.set_fixed_height(self.height)

    def close(self):
        with self.__draw_mutex:
            # if self.canvas_item:
            #     self.canvas_item.close()
            self.on_mouse_entered = None
            self.on_mouse_exited = None
            self.on_mouse_clicked = None
            self.on_mouse_double_clicked = None
            self.on_mouse_pressed = None
            self.on_mouse_released = None
            self.on_mouse_position_changed = None
            self.on_grabbed_mouse_position_changed = None
            self.on_wheel_changed = None
            self.on_key_pressed = None
            self.on_key_released = None
            self.on_size_changed = None
            self.on_drag_enter = None
            self.on_drag_leave = None
            self.on_drag_move = None
            self.on_drop = None
            self.on_tool_tip = None
            self.on_pan_gesture = None
            super().close()

    def _set_canvas_item(self, canvas_item):
        self.canvas_item.add_canvas_item(canvas_item)

    def periodic(self):
        if callable(self.on_periodic):
            self.on_periodic()

    @property
    def canvas_size(self):
        return self.height, self.width

    @property
    def focusable(self):
        return self.__focusable

    @focusable.setter
    def focusable(self, focusable):
        self.__focusable = focusable
        self._root_container.set_focusable(self, focusable)

    def draw(self, drawing_context):
        # thread safe. take care to make sure widget hasn't been deleted from underneath.
        with self.__draw_mutex:
            if self._root_container:
                self._root_container.draw(self.canvas_item, drawing_context)

    def set_cursor_shape(self, cursor_shape):
        pass

    def handle_mouse_entered(self):
        if callable(self.on_mouse_entered):
            self.on_mouse_entered()

    def handle_mouse_exited(self):
        if callable(self.on_mouse_exited):
            self.on_mouse_exited()

    def handle_mouse_clicked(self, x, y, modifiers):
        if callable(self.on_mouse_clicked):
            self.on_mouse_clicked(x, y, modifiers)

    def handle_mouse_double_clicked(self, x, y, modifiers):
        if callable(self.on_mouse_double_clicked):
            self.on_mouse_double_clicked(x, y, modifiers)

    def handle_mouse_pressed(self, x, y, modifiers):
        if callable(self.on_mouse_pressed):
            self.on_mouse_pressed(x, y, modifiers)

    def handle_mouse_released(self, x, y, modifiers):
        if callable(self.on_mouse_released):
            self.on_mouse_released(x, y, modifiers)

    def handle_mouse_position_changed(self, x, y, modifiers):
        if callable(self.on_mouse_position_changed):
            self.on_mouse_position_changed(x, y, modifiers)

    def handle_grabbed_mouse_position_changed(self, dx, dy, modifiers):
        if callable(self.on_grabbed_mouse_position_changed):
            self.on_grabbed_mouse_position_changed(dx, dy, modifiers)

    def handle_wheel_changed(self, x, y, dx, dy, is_horizontal):
        if callable(self.on_wheel_changed):
            self.on_wheel_changed(x, y, dx, dy, is_horizontal)

    def handle_size_changed(self, width, height):
        self.width = width
        self.height = height
        if self.width > 0:
            self.canvas_item.sizing.set_fixed_width(self.width)
        if self.height > 0:
            self.canvas_item.sizing.set_fixed_height(self.height)
        if callable(self.on_size_changed):
            self.on_size_changed(self.width, self.height)

    def handle_key_pressed(self, key):
        if callable(self.on_key_pressed):
            return self.on_key_pressed(key)
        return False

    def handle_key_released(self, key):
        if callable(self.on_key_released):
            return self.on_key_released(key)
        return False

    def handle_drag_enter_event(self, mime_data: UserInterface.MimeData) -> str:
        if callable(self.on_drag_enter):
            return self.on_drag_enter(mime_data)
        return "ignore"

    def handle_drag_leave_event(self):
        if callable(self.on_drag_leave):
            return self.on_drag_leave()
        return "ignore"

    def handle_drag_move_event(self, mime_data: UserInterface.MimeData, x: int, y: int) -> str:
        if callable(self.on_drag_move):
            return self.on_drag_move(mime_data, x, y)
        return "ignore"

    def handle_drop_event(self, mime_data: UserInterface.MimeData, x: int, y: int) -> str:
        if callable(self.on_drop):
            return self.on_drop(mime_data, x, y)
        return "ignore"

    def handle_tool_tip(self, x: int, y: int, gx: int, gy: int) -> bool:
        if callable(self.on_tool_tip):
            return self.on_tool_tip(x, y, gx, gy)
        return False

    def grab_gesture(self, gesture_type):
        self._root_container.grab_geture(self, gesture_type)

    def release_gesture(self, gesture_type):
        self._root_container.release_gesture(self, gesture_type)

    def handle_pan_gesture(self, delta_x, delta_y) -> bool:
        if callable(self.on_pan_gesture):
            return self.on_pan_gesture(delta_x, delta_y)
        return False


class Window(UserInterface.Window):

    def __init__(self, ui, parent_window, title):
        super().__init__(parent_window, title)
        self.ui = ui

    def close(self):
        super().close()

    # attach the root widget to this window
    # the root widget must either respond to _set_root_container or canvas_item
    def attach(self, root_widget):
        if not isinstance(root_widget, UserInterface.CanvasWidget):
            canvas_widget = self.ui.create_canvas_widget()
            canvas_widget.canvas_item.add_canvas_item(extract_canvas_item(root_widget))
            root_widget._set_root_container(self)
            root_widget = canvas_widget
        # root_widget should be a CanvasWidget
        super().attach(root_widget)
        size = self.__size
        if size is not None:
            self.root_widget._behavior.handle_size_changed(size.width, size.height)

    def _attach_root_widget(self, root_widget):
        pass

    # periodic is called periodically from the user interface object to service the window.
    def periodic(self):
        self._handle_periodic()

    # call show to display the window.
    def show(self, size=None, position=None):
        if self.on_about_to_show:
            self.on_about_to_show()

    def _set_title(self, value):
        pass

    def draw(self, canvas_item, drawing_context):
        """Render the drawing context by called draw on the ui object."""
        assert canvas_item == self.root_widget._behavior.canvas_item
        self.ui._draw(drawing_context)

    # called when the document size changes
    # typically by the UI object
    def handle_size_changed(self, size):
        self.__size = size
        if self.root_widget:
            if size is not None:
                self.root_widget._behavior.handle_size_changed(size.width, size.height)
        self._handle_size_changed(size.width, size.height)


class CanvasUserInterface(UserInterface.UserInterface):

    def __init__(self, draw_fn, get_font_metrics_fn):
        self.persistence_root = "0"
        self.__draw_fn = draw_fn
        self.__get_font_metrics_fn = get_font_metrics_fn
        self.__done = False
        self.__document_windows = list()

    def close(self):
        self.__done = True

    def run(self, event_queue):
        """Run the Python event loop.

        Wait for messages to arrive on event_queue and process them.
        """
        start = time.time()
        while not self.__done and time.time() - start < 1000.0:
            try:
                event_dict = event_queue.get(timeout=1/50.0)
                event_type = event_dict.get("type")
                document_window = self.__document_windows[0] if len(self.__document_windows) > 0 else None
                root_widget = document_window.root_widget if document_window else None

                if event_type == "quit":
                    return 0

                if root_widget:
                    if event_type == "mouse_enter":
                        root_widget._behavior.handle_mouse_entered()
                    elif event_type == "mouse_leave":
                        root_widget._behavior.handle_mouse_exited()
                    elif event_type == "mouse_down":
                        root_widget._behavior.handle_mouse_pressed(event_dict.get("x", 0.0), event_dict.get("y", 0.0), CanvasItem.KeyboardModifiers())
                    elif event_type == "mouse_up":
                        root_widget._behavior.handle_mouse_released(event_dict.get("x", 0.0), event_dict.get("y", 0.0), CanvasItem.KeyboardModifiers())
                    elif event_type == "mouse_move":
                        root_widget._behavior.handle_mouse_position_changed(event_dict.get("x", 0.0), event_dict.get("y", 0.0), CanvasItem.KeyboardModifiers())
                    elif event_type == "click":
                        root_widget._behavior.handle_mouse_clicked(event_dict.get("x", 0.0), event_dict.get("y", 0.0), CanvasItem.KeyboardModifiers())
                    elif event_type == "double_click":
                        root_widget._behavior.handle_mouse_double_clicked(event_dict.get("x", 0.0), event_dict.get("y", 0.0), CanvasItem.KeyboardModifiers())
                event_queue.task_done()
            except queue.Empty as e:
                pass
            try:
                for document_window in self.__document_windows:
                    document_window.periodic()
            except Exception as e:
                import traceback
                traceback.print_exc()
                traceback.print_stack()

    def _draw(self, drawing_context):
        """Render the drawing context.

        This will be called from the Window, which will, in turn, be called from a thread from its
        RootCanvasItem object."""
        dc = DrawingContext.DrawingContext()
        canvas_width = 960
        canvas_height = 720
        with dc.saver():
            dc.begin_path()
            dc.rect(0, 0, canvas_width, canvas_height)
            dc.fill_style = "#FFF"
            dc.fill()
        dc.add(drawing_context)
        self.__draw_fn(dc)

    # data objects

    def create_mime_data(self) -> UserInterface.MimeData:
        raise NotImplementedError()

    def create_item_model_controller(self, keys):
        raise NotImplementedError()

    def create_button_group(self):
        raise NotImplementedError()

    # window elements

    def create_document_window(self, title=None, parent_window=None):
        document_window = Window(self, parent_window, title)
        self.__document_windows.append(document_window)
        document_window.handle_size_changed(Geometry.IntSize(height=720, width=960))
        return document_window

    def destroy_document_window(self, document_window):
        if document_window in self.__document_windows:
            self.__document_windows.remove(document_window)

    # user interface elements

    def create_row_widget(self, alignment=None, properties=None):
        return UserInterface.BoxWidget(BoxWidgetBehavior("row", properties), alignment)

    def create_column_widget(self, alignment=None, properties=None):
        return UserInterface.BoxWidget(BoxWidgetBehavior("column", properties), alignment)

    def create_splitter_widget(self, orientation="vertical", properties=None):
        raise NotImplementedError()

    def create_tab_widget(self, properties=None):
        raise NotImplementedError()

    def create_stack_widget(self, properties=None):
        raise NotImplementedError()

    def create_scroll_area_widget(self, properties=None):
        raise NotImplementedError()

    def create_combo_box_widget(self, items=None, item_getter=None, properties=None):
        raise NotImplementedError()

    def create_push_button_widget(self, text=None, properties=None):
        return UserInterface.PushButtonWidget(PushButtonWidgetBehavior(self.__get_font_metrics_fn, properties), text)

    def create_radio_button_widget(self, text: str=None, properties=None):
        raise NotImplementedError()

    def create_check_box_widget(self, text=None, properties=None):
        return UserInterface.CheckBoxWidget(CheckBoxWidgetBehavior(self.__get_font_metrics_fn, properties), text)

    def create_label_widget(self, text=None, properties=None):
        return UserInterface.LabelWidget(LabelWidgetBehavior(self.get_font_metrics, properties), text)

    def create_slider_widget(self, properties=None):
        raise NotImplementedError()

    def create_progress_bar_widget(self, properties=None):
        raise NotImplementedError()

    def create_line_edit_widget(self, properties=None):
        raise NotImplementedError()

    def create_text_edit_widget(self, properties=None):
        raise NotImplementedError()

    def create_canvas_widget(self, properties=None):
        return UserInterface.CanvasWidget(CanvasWidgetBehavior(properties))

    def create_tree_widget(self, properties=None):
        raise NotImplementedError()

    # file i/o

    def load_rgba_data_from_file(self, filename):
        # returns data packed as uint32
        raise NotImplementedError()

    def save_rgba_data_to_file(self, data, filename, format):
        raise NotImplementedError()

    def get_existing_directory_dialog(self, title, directory):
        raise NotImplementedError()

    # persistence (associated with application)

    def get_data_location(self):
        raise NotImplementedError()

    def get_document_location(self):
        raise NotImplementedError()

    def get_temporary_location(self):
        raise NotImplementedError()

    def get_persistent_string(self, key, default_value=None):
        raise NotImplementedError()

    def set_persistent_string(self, key, value):
        raise NotImplementedError()

    def get_persistent_object(self, key, default_value=None):
        key = "/".join([self.persistence_root, key])
        value = self.get_persistent_string(key)
        return pickle.loads(binascii.unhexlify(value.encode("utf-8"))) if value else default_value

    def set_persistent_object(self, key, value):
        key = "/".join([self.persistence_root, key])
        self.set_persistent_string(key, binascii.hexlify(pickle.dumps(value, 0)).decode("utf-8"))

    def remove_persistent_key(self, key):
        raise NotImplementedError()

    # clipboard

    def clipboard_clear(self):
        raise NotImplementedError()

    def clipboard_mime_data(self) -> UserInterface.MimeData:
        raise NotImplementedError()

    def clipboard_set_mime_data(self, mime_data: UserInterface.MimeData) -> None:
        raise NotImplementedError()

    def clipboard_set_text(self, text):
        raise NotImplementedError()

    def clipboard_text(self):
        raise NotImplementedError()

    # misc

    def create_rgba_image(self, drawing_context, width, height):
        raise NotImplementedError()

    def get_font_metrics(self, font, text):
        return self.__get_font_metrics_fn(font, text)

    def create_context_menu(self, document_window):
        raise NotImplementedError()

    def create_sub_menu(self, document_window):
        raise NotImplementedError()
